# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.docstrings
import lale.operators
import torch
from pytorch_pretrained_bert import BertTokenizer, BertModel, BertForMaskedLM
import numpy as np
import pandas as pd
import logging
logging.basicConfig(level=logging.INFO)

class BertPretrainedEncoderImpl():
    def __init__(self, batch_size = 32):
        # Load pre-trained model tokenizer (vocabulary)
        self.tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
        self.max_seq_length = self.tokenizer.max_len
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model = BertModel.from_pretrained('bert-base-uncased')
        self.batch_size = batch_size

    # def fit(self, X, y):
    #     # TODO: Find the right value for max sequence length
    #     return BertPretrainedEncoderImpl()

    def transform(self, X):
        if isinstance(X, np.ndarray) or isinstance(X, pd.DataFrame):
            X = X.squeeze()
        self.model.eval()
        self.model.to(self.device)
        transformed_X = None
        num_batches = len(X)//self.batch_size
        last_batch_size = len(X)%self.batch_size
        for batch_idx in range(num_batches+1):
          min_idx = batch_idx*self.batch_size
          if batch_idx+1 <= num_batches:
            max_idx = (batch_idx+1)*self.batch_size            
          else:
            max_idx = batch_idx*self.batch_size + last_batch_size
          batch_data = X[min_idx:max_idx]
          indexed_tokenized_X = []
          segments_ids = []
          # Convert token to vocabulary indices
          for line in batch_data:
              tokenized_text = self.tokenizer.tokenize(f'[CLS] {line} [SEP]')
              tokenized_ids = self.tokenizer.convert_tokens_to_ids(tokenized_text)
              if len(tokenized_ids) > self.max_seq_length - 2:
                  tokenized_ids = tokenized_ids[:(self.max_seq_length)]            

              # The mask has 1 for real tokens and 0 for padding tokens. Only real
              # tokens are attended to.
              input_mask = [1] * len(tokenized_ids)
              # Zero-pad up to the sequence length.
              padding = [0] * (self.max_seq_length - len(tokenized_ids))
              padded_tokenized_ids = tokenized_ids + padding
              input_mask += padding            
              indexed_tokenized_X.append(padded_tokenized_ids)

              #This transformer is only applicable for single sentences as we are passing all
              # segment ids as 1. BERT has a notion of a sentence pair, say for a question-answering task
              # that would need a different handling of segments_ids
              segments_ids.append([1 for token in padded_tokenized_ids])

          # Convert inputs to PyTorch tensors
          tokens_tensor = torch.tensor(indexed_tokenized_X)
          segments_tensors = torch.tensor(segments_ids)

          # Predict hidden states features for each layer
          with torch.no_grad():
            encoded_layers, _ = self.model(tokens_tensor.to(self.device), segments_tensors.to(self.device))
          batch_X = encoded_layers[-1][:, 0, :].detach().cpu().numpy()
          if transformed_X is None:
            transformed_X = batch_X
          else:
            transformed_X = np.vstack((transformed_X, batch_X))
        return transformed_X



_input_schema_fit = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Input data schema for training.',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'anyOf': [
        { 'type': 'array',
          'items': {'type': 'string'}},
        { 'type': 'array',
          'items': {
            'type': 'array', 'minItems': 1, 'maxItems': 1,
            'items': {'type': 'string'}}}]},
    'y': {
      'description': 'Labels, optional.'
    }
    }}

_input_transform_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Input data schema for predictions.',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'anyOf': [
        { 'type': 'array',
          'items': {'type': 'string'}},
        { 'type': 'array',
          'items': {
            'type': 'array', 'minItems': 1, 'maxItems': 1,
            'items': {'type': 'string'}}}]}}}

_output_transform_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Output data schema for transformed data.',
  'type': 'array',
  'items': {'type': 'array', 'items': {'type': 'number'}}}
  

_hyperparams_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Hyperparameter schema.',
  'allOf': [
    { 'description':
        'This first sub-object lists all constructor arguments with their '
        'types, one at a time, omitting cross-argument constraints.',
      'type': 'object',
      'additionalProperties': False,
      'relevantToOptimizer': [],
      'properties': {
        'batch_size':{
          'description': 'Batch size used for transform.',
          'type': 'integer',
          'default': 64,
          'minimum': 1,
          'distribution': 'uniform',
          'minimumForOptimizer': 32,
          'maximumForOptimizer': 128}}}]}

_combined_schemas = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Combined schema for expected data and hyperparameters for a transformer for'
                ' a text data transformer based on pre-trained BERT model '
                '(https://github.com/huggingface/pytorch-pretrained-BERT).',
  'type': 'object',
  'tags': {
    'pre': ['text'],
    'op': ['transformer', '~interpretable'],
    'post': ['embedding']},
  'properties': {
    'hyperparams': _hyperparams_schema,
    'input_fit': _input_schema_fit,
    'input_transform': _input_transform_schema,
    'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(BertPretrainedEncoderImpl, _combined_schemas)

BertPretrainedEncoder = lale.operators.make_operator(BertPretrainedEncoderImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Iterable, Optional, List, Set, Tuple, TypeVar, Union
import random
import math
import warnings

import lale.operators as Ops
from lale.pretty_print import hyperparams_to_string
from lale.search.PGO import remove_defaults_dict
from lale.util.Visitor import Visitor

# We support an argument encoding schema intended to be a 
# conservative extension to sklearn's encoding schema
# sklearn uses __ to separate elements in a hierarchy
# (pipeline's have operators that have keys)

# Since we support richer computation graphs, we need to extend this encoding
# to support it. Graphs that could be represented in sklearn
# should be encoded identically

# our encoding scheme:
## __ separates nested components (as-in sklearn)
## ? is the discriminant (choice made) for a choice
## ? is also a prefix for the nested parts of the chosen branch
## x@n In a pipeline, if multiple components have identical names,
### everything but the first are suffixed with a number (starting with 1)
### indicating which one we are talking about.
### For example, given (x >> y >> x), we would treat this much the same as
### (x >> y >> x@1)
## $ is used in the rare case that sklearn would expect the key of an object,
### but we allow (and have) a non-object schema.  In that case,
### $ is used as the key. This should only happen at the top level,
### since nested occurences should be removed.
## # is a structure indicator, and the value should be one of 'list', 'tuple', or 'dict'
## n is used to represent the nth component in an array or tuple

# This method (and the to_lale() method on the returned value)
# are the only ones intended to be exported
def make_sklearn_compat(op:Ops.Operator)->'SKlearnCompatWrapper':
    """Top level function for providing compatibiltiy with sklearn operations
       This returns a wrapper around the provided sklearn operator graph which can be passed
       to sklearn methods such as clone and GridSearchCV
       The wrapper may modify the wrapped lale operator/pipeline as part of providing
       compatibility with these methods.
       After the sklearn operation is complete, 
       SKlearnCompatWrapper.to_lale() can be called to recover the 
       wrapped lale operator for future use
    """
    return SKlearnCompatWrapper.make_wrapper(op)

def sklearn_compat_clone(impl:Any)->Any:
    if impl is None:
        return None
    
    from sklearn.base import clone
    cp = clone(impl, safe=False)
    return cp

def clone_lale(op:Ops.Operator)->Ops.Operator:
    return op._lale_clone(sklearn_compat_clone)

class WithoutGetParams(object):
    """ This wrapper forwards everything except "get_attr" to what it is wrapping
    """
    def __init__(self, base):
        self._base = base
        assert self._base != self

    def __getattr__(self, name):
        # This is needed because in python copy skips calling the __init__ method
        if name == "_base":
            raise AttributeError
        if name == 'get_params':
            raise AttributeError
        if name in ['__getstate__', '__setstate__', '__repr__']:
            raise AttributeError
        else:
            return getattr(self._base, name)
    
    @classmethod
    def clone_wgp(cls, obj:'WithoutGetParams')->'WithoutGetParams':
        while isinstance(obj, WithoutGetParams):
            obj = obj._base
        assert isinstance(obj, Ops.Operator)
        return WithoutGetParams(clone_lale(obj))

    def __str__(self):
        b = getattr(self, '_base', None)
        s:str
        if b is None:
            s = ""
        else:
            s = str(b)       
        return f"WGP<{s}>"

def partition_sklearn_params(d:Dict[str, Any])->Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
    sub_parts:Dict[str, Dict[str, Any]] = {}
    main_parts:Dict[str,Any] = {}
    
    for k, v in d.items():
        ks = k.split("__", 1)
        if len(ks) == 1:
            assert k not in main_parts
            main_parts[k] = v
        else:
            assert len(ks) == 2
            bucket:Dict[str, Any] = {}
            group:str = ks[0]
            param:str = ks[1]
            if group in sub_parts:
                bucket = sub_parts[group]
            else:
                sub_parts[group] = bucket
            assert param not in bucket
            bucket[param] = v
    return (main_parts, sub_parts)

def partition_sklearn_choice_params(d:Dict[str, Any])->Tuple[int, Dict[str, Any]]:
    discriminant_value:int = -1
    choice_parts:Dict[str,Any] = {}

    for k, v in d.items():
        if k == discriminant_name:
            assert discriminant_value == -1
            discriminant_value = int(v)
        else:
            k_rest = unnest_choice(k)
            choice_parts[k_rest] = v
    assert discriminant_value != -1
    return (discriminant_value, choice_parts)

DUMMY_SEARCH_SPACE_GRID_PARAM_NAME:str = "$"
discriminant_name:str = "?"
choice_prefix:str = "?"
structure_type_name:str = "#"
structure_type_list:str = "list"
structure_type_tuple:str = "tuple"
structure_type_dict:str = "dict"


def get_name_and_index(name:str)->Tuple[str,int]:
    """ given a name of the form "name@i", returns (name, i)
        if given a name of the form "name", returns (name, 0)
    """
    splits = name.split("@",1)
    if len(splits) == 1:
        return splits[0], 0
    else:
        return splits[0], int(splits[1])

def make_degen_indexed_name(name, index):
    return f"{name}@{index}"

def make_indexed_name(name, index):
    if index == 0:
        return name
    else:
        return f"{name}@{index}"

def make_array_index_name(index, is_tuple:bool=False):
    sep = "##" if is_tuple else "#"
    return f"{sep}{str(index)}"

def is_numeric_structure(structure_type:str):

    if structure_type == "list" or structure_type == "tuple":
        return True
    elif structure_type == "dict":
        return False
    else:
        assert False, f"Unknown structure type {structure_type} found"

def set_structured_params(k, params:Dict[str, Any], hyper_parent):
    # need to handle the different encoding schemes used
    if params is None:
        return None
    if structure_type_name in params:
        # this is a structured type
        structure_type = params[structure_type_name]
        type_params, sub_params = partition_sklearn_params(params)

        hyper = None
        if isinstance(hyper_parent, dict):
            hyper = hyper_parent.get(k, None)
        elif isinstance(hyper_parent, list) and k < len(hyper_parent):
            hyper = hyper_parent[k]
        if hyper is None:
                hyper = {}
        elif isinstance(hyper, tuple):
            # to make it mutable
            hyper = list(hyper)

        del type_params[structure_type_name]
        actual_key:Union[str,int]
        for elem_key, elem_value in type_params.items():
            if elem_value is not None:
                if not isinstance(hyper, dict):
                    assert is_numeric_structure(structure_type)
                    actual_key = int(elem_key)
                else:
                    actual_key = elem_key
                hyper[actual_key] = elem_value

        for elem_key, elem_params in sub_params.items():
            if not isinstance(hyper, dict):
                assert is_numeric_structure(structure_type)
                actual_key = int(elem_key)
            else:
                actual_key = elem_key
            set_structured_params(actual_key, elem_params, hyper)
        if isinstance(hyper, dict) and is_numeric_structure(structure_type):
            max_key = max(map(int,hyper.keys()))
            hyper = [hyper.get(str(x), None) for x in range(max_key)]
        if structure_type == "tuple":
            hyper = tuple(hyper)
        hyper_parent[k] = hyper
    else:
        # if it is not a structured parameter
        # then it must be a nested higher order operator
        sub_op = hyper_parent[k]
        if isinstance(sub_op, list):
            if len(sub_op) == 1:
                sub_op = sub_op[0]
            else:
                (disc, chosen_params) = partition_sklearn_choice_params(params)
                assert 0 <= disc and disc < len(sub_op)
                sub_op = sub_op[disc]
                params = chosen_params
        trainable_sub_op = set_operator_params(sub_op, **params)
        hyper_parent[k] = trainable_sub_op

def set_operator_params(op:Ops.Operator, **impl_params)->Ops.TrainableOperator:
    """May return a new operator, in which case the old one should be overwritten
    """
    if isinstance(op, Ops.PlannedIndividualOp):
        main_params, partitioned_sub_params = partition_sklearn_params(impl_params)
        hyper = op._hyperparams
        if hyper is None:
            hyper = {}
        # we set the sub params first
        for sub_key, sub_params in partitioned_sub_params.items():
            set_structured_params(sub_key, sub_params, hyper)

         # we have now updated any nested operators
         # (if this is a higher order operator)
         # and can work on the main operator
        all_params = {**main_params, **hyper}
        return op.set_params(**all_params)
    elif isinstance(op, Ops.BasePipeline):
        steps = op.steps()
        main_params, partitioned_sub_params = partition_sklearn_params(impl_params)
        assert not main_params, f"Unexpected non-nested arguments {main_params}"
        found_names:Dict[str, int] = {}
        step_map:Dict[Ops.Operator, Ops.TrainableOperator] = {}
        for s in steps:
            name = s.name()
            name_index = 0
            params:Dict[str, Any] = {}
            if name in found_names:
                name_index = found_names[name] + 1
                found_names[name] = name_index
                uname = make_indexed_name(name, name_index)
                if uname  in partitioned_sub_params:
                    params = partitioned_sub_params[uname]
            else:
                found_names[name] = 0
                uname = make_degen_indexed_name(name, 0)
                if uname in partitioned_sub_params:
                    params = partitioned_sub_params[uname]
                    assert name not in partitioned_sub_params
                elif name in partitioned_sub_params:
                    params = partitioned_sub_params[name]
            new_s = set_operator_params(s, **params)
            if s != new_s:
                step_map[s] = new_s
        # make sure that no parameters were passed in for operations
        # that are not actually part of this pipeline
        for k in partitioned_sub_params.keys():
            n,i = get_name_and_index(k)
            assert n in found_names and i <= found_names[n]
        if step_map:
            op.subst_steps(step_map)
            if not isinstance(op, Ops.TrainablePipeline):
                # As a result of choices made, we may now be a TrainableIndividualOp
                ret = Ops.get_pipeline_of_applicable_type(op.steps(), op.edges(), ordered=True)
                if not isinstance(ret, Ops.TrainableOperator):
                    assert False
                return ret
            else:
                return op
        else:
            assert isinstance(op, Ops.TrainableOperator)
            return op
    elif isinstance(op, Ops.OperatorChoice):
        choices = op.steps()
        choice_index:int
        choice_params:Dict[str, Any]
        if len(choices)==1:
            choice_index = 0
            chosen_params = impl_params
        else:
            (choice_index, chosen_params) = partition_sklearn_choice_params(impl_params)

        assert 0 <= choice_index and choice_index < len(choices)
        choice:Ops.Operator = choices[choice_index]

        new_step = set_operator_params(choice, **chosen_params)
        # we remove the OperatorChoice, replacing it with the branch that was taken
        return new_step
    else:
        assert False, f"Not yet supported operation of type: {op.__class__.__name__}"

class SKlearnCompatWrapper(object):
    _base:WithoutGetParams
    # This is used to trick clone into leaving us alone
    _old_params_for_clone:Optional[Dict[str, Any]]

    @classmethod
    def make_wrapper(cls, base:Ops.Operator):
        b:Any = base
        if isinstance(base, SKlearnCompatWrapper):
            return base
        elif not isinstance(base, WithoutGetParams):
            b = WithoutGetParams(base)
        return cls(__lale_wrapper_init_base=b)

    def __init__(self, **kwargs):
        if '__lale_wrapper_init_base' in kwargs:
            # if we are being called by make_wrapper
            # then we don't need to make a copy
            self._base = kwargs['__lale_wrapper_init_base']
            self._old_params_for_clone = None

        else:
            # otherwise, we are part of a get_params/init clone
            # and we need to make a copy
            self.init_params_internal(**kwargs)
        assert self._base != self

    def init_params_internal(self, **kwargs):
        op = kwargs['__lale_wrapper_base']
        self._base = WithoutGetParams.clone_wgp(op)
        self._old_params_for_clone = kwargs

    def get_params_internal(self, out:Dict[str,Any]):
        out['__lale_wrapper_base'] = self._base

    def set_params_internal(self, **impl_params):
        self._base = impl_params['__lale_wrapper_base']
        assert self._base != self

    def fixup_params_internal(self, **params):
        return params

    def to_lale(self)->Ops.Operator:
        cur:Any = self
        assert cur != None
        assert cur._base != None
        cur = cur._base
        while isinstance(cur, WithoutGetParams):
            cur = cur._base
        assert isinstance(cur, Ops.Operator)
        return cur

    # sklearn calls __repr__ instead of __str__
    def __repr__(self):
        op = self.to_lale()
        if isinstance(op, Ops.TrainableIndividualOp):
            name = op.name()
            hyps = ""
            hps = op.hyperparams()
            if hps is not None:
                hyps = hyperparams_to_string(hps)
            return name + "(" + hyps + ")"
        else:
            return super().__repr__()

    def __getattribute__(self, name):
        """ Try proxying unknown attributes to the underlying operator
            getattribute is used instead of getattr to ensure that the 
            correct underlying error is thrown in case
            a property (such as classes_) throws an AttributeError
        """

        # This is needed because in python copy skips calling the __init__ method
        try:
            return super(SKlearnCompatWrapper, self).__getattribute__(name)
        except AttributeError as e:
            if name == "_base":
                raise AttributeError
            try:
                return getattr(self._base, name)
            except AttributeError:
                raise e



    def get_params(self, deep:bool = True)->Dict[str,Any]:
        out:Dict[str,Any] = {}
        if not deep:
            if self._old_params_for_clone is not None:
                # lie to clone to make it happy
                params = self._old_params_for_clone
                self._old_params_for_clone = None
                return params
            else:
                self.get_params_internal(out)
        else:
            pass #TODO
        return out

    def fit(self, X, y=None, **fit_params):
        if hasattr(self._base, 'fit'):
            filtered_params = remove_defaults_dict(fit_params)
            return self._base.fit(X, y=y, **filtered_params)
        else:
            pass

    def set_params(self, **impl_params):

        if '__lale_wrapper_base' in impl_params:
            self.set_params_internal(**impl_params)
        else:
            prev = self
            cur = self._base
            assert prev != cur
            assert cur != None
            while isinstance(cur, WithoutGetParams):
                assert cur != cur._base
                prev = cur
                cur = cur._base
            if not isinstance(cur, Ops.Operator):
                assert False
            assert isinstance(cur, Ops.Operator)
            fixed_params = self.fixup_params_internal(**impl_params)
            new_s = set_operator_params(cur, **fixed_params)
            if not isinstance(new_s, Ops.TrainableOperator):
                assert False
            if new_s != cur:
                prev._base = new_s                
        return self

    def hyperparam_defaults(self)->Dict[str,Any]:
        return DefaultsVisitor.run(self.to_lale())

    def _final_individual_op(self)->Optional[Ops.IndividualOp]:
        op:Optional[Ops.Operator] = self.to_lale()
        while op is not None and isinstance(op, Ops.BasePipeline):
            op = op.get_last()
        if op is not None and not isinstance(op, Ops.IndividualOp):
            op = None
        return op

    @property
    def _final_estimator(self):
        op:Optional[Ops.IndividualOp] = self._final_individual_op()
        model = None
        if op is not None:
            # if fit was called, we want to use trained result
            # even if the code uses the original operrator
            # since sklearn assumes that fit mutates the operator
            if hasattr(op, '_trained'):
                op = op._trained
            if hasattr(op, '_impl') and hasattr(op._impl_instance(), '_wrapped_model'):
                model = op._impl_instance()._wrapped_model
        return 'passthrough' if model is None else model

    @property
    def classes_(self):
        return self._final_estimator.classes_

    @property
    def n_classes_(self):
        return self._final_estimator.n_classes_

    @property
    def _estimator_type(self):
        return self._final_estimator._estimator_type

    # sklearn compatibility
    # @property
    # def _final_estimator(self):
    #     lale_op = self.to_lale()
    #     if lale_op is _

    #     estimator = self.steps[-1][1]
    #     return 'passthrough' if estimator is None else estimator
        
class DefaultsVisitor(Visitor):
    @classmethod
    def run(cls, op:Ops.Operator)->Dict[str,Any]:
        visitor = cls()
        accepting_op:Any = op
        return accepting_op.accept(visitor)

    def __init__(self):
        super(DefaultsVisitor, self).__init__()

    def visitIndividualOp(self, op:Ops.IndividualOp)->Dict[str,Any]:
        return op.hyperparam_defaults()

    visitPlannedIndividualOp = visitIndividualOp
    visitTrainableIndividualOp = visitIndividualOp
    visitTrainedIndividualOp = visitIndividualOp

    def visitPipeline(self, op:Ops.PlannedPipeline)->Dict[str,Any]:

        defaults_list:Iterable[Dict[str,Any]] = (
            nest_HPparams(s.name(), s.accept(self)) for s in op.steps())

        defaults:Dict[str,Any] = {}
        for d in defaults_list:
            defaults.update(d)

        return defaults

    visitPlannedPipeline = visitPipeline
    visitTrainablePipeline = visitPipeline
    visitTrainedPipeline = visitPipeline

    def visitOperatorChoice(self, op:Ops.OperatorChoice)->Dict[str,Any]:

        defaults_list:Iterable[Dict[str,Any]] = (
            s.accept(self) for s in op.steps())

        defaults : Dict[str,Any] = {}
        for d in defaults_list:
            defaults.update(d)

        return defaults

# Auxiliary functions
V = TypeVar('V')

def nest_HPparam(name:str, key:str):
    if key == DUMMY_SEARCH_SPACE_GRID_PARAM_NAME:
        # we can get rid of the dummy now, since we have a name for it
        return name
    return name + "__" + key

def nest_HPparams(name:str, grid:Dict[str,V])->Dict[str,V]:
    return {(nest_HPparam(name, k)):v for k, v in grid.items()}

def nest_all_HPparams(name:str, grids:List[Dict[str,V]])->List[Dict[str,V]]:
    """ Given the name of an operator in a pipeline, this transforms every key(parameter name) in the grids
        to use the operator name as a prefix (separated by __).  This is the convention in scikit-learn pipelines.
    """
    return [nest_HPparams(name, grid) for grid in grids]

def nest_choice_HPparam(key:str):
    return choice_prefix + key

def nest_choice_HPparams(grid:Dict[str,V])->Dict[str,V]:
    return {(nest_choice_HPparam(k)):v for k, v in grid.items()}

def nest_choice_all_HPparams(grids:List[Dict[str,V]])->List[Dict[str,V]]:
    """ this transforms every key(parameter name) in the grids
        to be nested under a choice, using a ? as a prefix (separated by __).  This is the convention in scikit-learn pipelines.
    """
    return [nest_choice_HPparams(grid) for grid in grids]

def unnest_choice(k:str)->str:
    assert k.startswith(choice_prefix)
    return k[len(choice_prefix):]

def unnest_HPparams(k:str)->List[str]:
    return k.split("__")    

OpType = TypeVar('OpType', bound=Ops.Operator)
def clone_op(op: OpType, name:str=None) -> OpType:
    """ Clone any operator.
    """
    from sklearn.base import clone
    nop = clone(make_sklearn_compat(op)).to_lale()
    if name:
        nop.set_name(name)
    return nop




from sklearn.decomposition.sparse_pca import SparsePCA as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class SparsePCAImpl():

    def __init__(self, n_components=None, alpha=1, ridge_alpha=0.01, max_iter=1000, tol=1e-08, method='lars', n_jobs=None, U_init=None, V_init=None, verbose=False, random_state=None, normalize_components=False):
        self._hyperparams = {
            'n_components': n_components,
            'alpha': alpha,
            'ridge_alpha': ridge_alpha,
            'max_iter': max_iter,
            'tol': tol,
            'method': method,
            'n_jobs': n_jobs,
            'U_init': U_init,
            'V_init': V_init,
            'verbose': verbose,
            'random_state': random_state,
            'normalize_components': normalize_components}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for SparsePCA    Sparse Principal Components Analysis (SparsePCA)',
    'allOf': [{
        'type': 'object',
        'required': ['n_components', 'alpha', 'ridge_alpha', 'max_iter', 'tol', 'method', 'n_jobs', 'U_init', 'V_init', 'verbose', 'random_state', 'normalize_components'],
        'relevantToOptimizer': ['n_components', 'alpha', 'max_iter', 'tol', 'method'],
        'additionalProperties': False,
        'properties': {
            'n_components': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 2,
                    'maximumForOptimizer': 256,
                    'distribution': 'uniform'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Number of sparse atoms to extract.'},
            'alpha': {
                'type': 'number',
                'minimumForOptimizer': 1e-10,
                'maximumForOptimizer': 1.0,
                'distribution': 'uniform',
                'default': 1,
                'description': 'Sparsity controlling parameter'},
            'ridge_alpha': {
                'type': 'number',
                'default': 0.01,
                'description': 'Amount of ridge shrinkage to apply in order to improve conditioning when calling the transform method.'},
            'max_iter': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 1000,
                'description': 'Maximum number of iterations to perform.'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'uniform',
                'default': 1e-08,
                'description': 'Tolerance for the stopping condition.'},
            'method': {
                'enum': ['lars', 'cd'],
                'default': 'lars',
                'description': 'lars: uses the least angle regression method to solve the lasso problem (linear_model.lars_path) cd: uses the coordinate descent method to compute the Lasso solution (linear_model.Lasso)'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Number of parallel jobs to run'},
            'U_init': {
                'anyOf': [{
                    'type': 'array',
                    'items': {
                        'type': 'array',
                        'items': {
                            'type': 'number'},
                    }}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Initial values for the loadings for warm restart scenarios.'},
            'V_init': {
                'anyOf': [{
                    'type': 'array',
                    'items': {
                        'type': 'array',
                        'items': {
                            'type': 'number'},
                    }}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Initial values for the components for warm restart scenarios.'},
            'verbose': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'boolean'}],
                'default': False,
                'description': 'Controls the verbosity; the higher, the more messages'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`.'},
            'normalize_components': {
                'type': 'boolean',
                'default': False,
                'description': '- if False, use a version of Sparse PCA without components   normalization and without data centering'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the model from data in X.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vector, where n_samples in the number of samples and n_features is the number of features.'},
        'y': {
            
        }},
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Least Squares projection of the data onto the sparse components.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Test data to be transformed, must have the same number of features as the data used to train the model.'},
        'ridge_alpha': {
            'type': 'number',
            'default': 0.01,
            'description': 'Amount of ridge shrinkage to apply in order to improve conditioning'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transformed data.',
    'laleType': 'Any',
    'XXX TODO XXX': 'X_new array, shape (n_samples, n_components)',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.decomposition.SparsePCA#sklearn-decomposition-sparsepca',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(SparsePCAImpl, _combined_schemas)
SparsePCA = lale.operators.make_operator(SparsePCAImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.docstrings
import lale.operators
import numpy as np
import pandas as pd
import sklearn.feature_extraction.text

class TfidfVectorizerImpl():
    def __init__(self, **hyperparams):
        if 'dtype' in hyperparams and hyperparams['dtype'] == 'float64':
            hyperparams = {**hyperparams, 'dtype': np.float64}
        self._hyperparams = hyperparams
        self._wrapped_model = sklearn.feature_extraction.text.TfidfVectorizer(**self._hyperparams)

    def fit(self, X, y=None):
        if isinstance(X, np.ndarray) or isinstance(X, pd.DataFrame):
            X = X.squeeze()
        self._wrapped_model.fit(X, y)
        return self

    def transform(self, X):
        if isinstance(X, np.ndarray) or isinstance(X, pd.DataFrame):
            X = X.squeeze()
        return self._wrapped_model.transform(X)

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Convert a collection of raw documents to a matrix of TF-IDF features.',
    'allOf': [
    {   'type': 'object',
        'required': ['input', 'encoding', 'decode_error', 'strip_accents', 'lowercase', 'preprocessor', 'tokenizer', 'analyzer', 'stop_words', 'ngram_range', 'max_df', 'min_df', 'max_features', 'vocabulary', 'binary', 'dtype', 'norm', 'use_idf', 'smooth_idf', 'sublinear_tf'],
        'relevantToOptimizer': ['analyzer',  'ngram_range', 'max_df', 'min_df', 'binary', 'norm', 'use_idf', 'smooth_idf', 'sublinear_tf'],
        'additionalProperties': False,
        'properties': {
            'input': {
                'enum': ['filename', 'file', 'content'],
                'default': 'content'},
            'encoding': {
                'type': 'string',
                'default': 'utf-8'},
            'decode_error': {
                'enum': ['strict', 'ignore', 'replace'],
                'default': 'strict'},
            'strip_accents': {
                'enum': ['ascii', 'unicode', None],
                'default': None},
            'lowercase': {
                'type': 'boolean',
                'default': True},
            'preprocessor': {
                'anyOf': [
                {},
                {   'enum': [None]}],
                'default': None},
            'tokenizer': {
                'anyOf': [
                {},
                {   'enum': [None]}],
                'default': None},
            'analyzer': {
                'enum': ['word', 'char', 'char_wb'], #Need to support callable
                'default': 'word'},
            'stop_words': {
                'anyOf': [{
                    'enum': [None]}, {
                    'type': 'array'}, {
                    'type': 'string'}],
                'default': None},
            'token_pattern': {
                'type': 'string',
                'default': '(?u)\\b\\w\\w+\\b'},
            'ngram_range': {
                'default': [1, 1],
                'anyOf': [{
                    'type': 'array',
                    'laleType': 'tuple',
                    'minItemsForOptimizer': 2,
                    'maxItemsForOptimizer': 2,
                    'items': {
                        'type': 'integer',
                        'minimumForOptimizer': 1,
                        'maximumForOptimizer': 3},
                    'forOptimizer':False
                    },
                    {
                        'enum': [(1,1), (1,2), (1,3), (2,2), (2,3), (3,3)]
                    }
                ]},
            'max_df': {
                'anyOf': [{
                    'description': 'float in range [0.0, 1.0]',
                    'type': 'number',
                    'minimum': 0.0,
                    'maximum': 1.0,
                    'minimumForOptimizer': 0.8,
                    'maximumForOptimizer': 0.9,
                    'distribution': 'uniform'}, {
                    'type': 'integer', 'forOptimizer': False}],
                'default': 1.0},
            'min_df': {
                'anyOf': [{
                    'description': 'float in range [0.0, 1.0]',
                    'type': 'number',
                    'minimum': 0.0,
                    'maximum': 1.0,
                    'minimumForOptimizer': 0.0,
                    'maximumForOptimizer': 0.1,
                    'distribution': 'uniform'}, {
                    'type': 'integer', 'forOptimizer': False}],
                'default': 1},
            'max_features': {
                'anyOf': [{
                    'anyOf': [{
                        'type': 'number'}, {
                        'enum': [None]}]}],
                'default': None},
            'vocabulary': {
                'description': 'XXX TODO XXX, Mapping or iterable, optional',
                'anyOf': [{
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None},
            'binary': {
                'type': 'boolean',
                'default': False},
            'dtype': {
                'description': 'XXX TODO XXX, type, optional',
                'type': 'string',
                'default': 'float64'},
            'norm': {
                'enum': ['l1', 'l2', None],
                'default': 'l2'},
            'use_idf': {
                'type': 'boolean',
                'default': True},
            'smooth_idf': {
                'type': 'boolean',
                'default': True},
            'sublinear_tf': {
                'type': 'boolean',
                'default': False}}},
    {   'description': "tokenizer, only applies if analyzer == 'word'",
        'anyOf': [
        {   'type': 'object',
            'properties': {
                'analyzer': {
                    'enum': ['word']}}},
        {   'type': 'object',
            'properties': {
                'tokenizer': {
                    'enum': [None]}}}]},
    {   'description': "stop_words can be a list only if analyzer == 'word'",
        'anyOf': [
        {   'type': 'object',
            'properties': {
                'stop_words': {
                    'not': {
                        'type': 'array'}}}},
        {   'type': 'object',
            'properties': {
                'analyzer': {
                    'enum': ['word']}}}]}]}

_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for training the TfidfVectorizer from scikit-learn.',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Features; the outer array is over samples.',
            'anyOf': [
                {   'type': 'array',
                    'items': {'type': 'string'}},
                {   'type': 'array',
                    'items': {
                        'type': 'array', 'minItems': 1, 'maxItems': 1,
                        'items': {'type': 'string'}}}]},
        'y': {
            'description': 'Target class labels; the array is over samples.'}}}

_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for predictions using the TfidfVectorizer model from scikit-learn.',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Features; the outer array is over samples.',
            'anyOf': [
                {   'type': 'array',
                    'items': {'type': 'string'}},
                {   'type': 'array',
                    'items': {
                        'type': 'array', 'minItems': 1, 'maxItems': 1,
                        'items': {'type': 'string'}}}]}}}

_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Output data schema for predictions (projected data) using the TfidfVectorizer model from scikit-learn.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'}}}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html#sklearn-feature-extraction-text-tfidfvectorizer',
    'type': 'object',
    'tags': {
        'pre': ['text'],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(TfidfVectorizerImpl, _combined_schemas)

TfidfVectorizer = lale.operators.make_operator(TfidfVectorizerImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from torch.utils.data import DataLoader
import copy
import torchvision.transforms as transforms
from torchvision.models import resnet50
import torch.optim as optim
import torch.nn as nn
import torch.nn.functional as F

from torch.autograd import Variable
import sys
import numpy as np
import math
import lale.operators
import lale.docstrings

class ResNet50Impl():
    def __init__(self, num_classes=10, model = None, 
            num_epochs = 2, batch_size = 128, learning_rate_init=0.1,
            learning_rate = 'constant'):
        self.num_classes = num_classes
        if model is None:
            #self.model = ResNet(152, num_classes)
            self.model = resnet50(num_classes)
        else:
            self.model = model
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.num_epochs = num_epochs
        self.batch_size = batch_size
        self.learning_rate_init = learning_rate_init
        self.learning_rate = learning_rate

    def calculate_learning_rate(self, epoch):
        if self.learning_rate == 'constant':
            return self.learning_rate_init
        elif self.learning_rate == 'decay':
            optim_factor = 0
            if(epoch > 160):
                optim_factor = 3
            elif(epoch > 120):
                optim_factor = 2
            elif(epoch > 60):
                optim_factor = 1

            return self.learning_rate_init*math.pow(0.2, optim_factor)

    def fit(self, X, y = None):
        trainloader = torch.utils.data.DataLoader(X, batch_size=self.batch_size, shuffle=True, num_workers=2)
        net = self.model.to(self.device)
        net.train()
        train_loss = 0
        correct = 0
        total = 0
        for epoch in range(self.num_epochs):
            learning_rate = self.calculate_learning_rate(epoch)
            print(learning_rate)
            optimizer = optim.SGD(net.parameters(), lr=learning_rate, momentum=0.9, weight_decay=5e-4)
            criterion = nn.CrossEntropyLoss()
            print('\n=> Training Epoch %d, LR=%.4f' %(epoch, learning_rate))
            for batch_idx, (inputs, targets) in enumerate(trainloader):
                inputs, targets = inputs.to(self.device), targets.to(self.device) # GPU settings
                optimizer.zero_grad()
                inputs, targets = Variable(inputs), Variable(targets)
                outputs = net(inputs)               # Forward Propagation
                loss = criterion(outputs, targets)  # Loss
                loss.backward()  # Backward Propagation
                optimizer.step() # Optimizer update

                train_loss += loss.data.item()
                _, predicted = torch.max(outputs.data, 1)
                total += targets.size(0)
                correct += predicted.eq(targets.data).cpu().sum()

                sys.stdout.write('\r')
                sys.stdout.write('| Epoch [%3d/%3d] Iter[%3d/%3d]\t\tLoss: %.4f Acc@1: %.3f%%'
                        %(epoch, self.num_epochs, batch_idx+1,
                            (len(X)//self.batch_size)+1, loss.data.item(), 100.*correct/total))
                sys.stdout.flush()
        return ResNet50Impl(self.num_classes, self.model, self.num_epochs, self.batch_size,
                            self.learning_rate) 

    def predict(self, X):
        net = self.model.to(self.device)
        if torch.cuda.device_count() > 1:
            net = nn.DataParallel(net)
        net.eval()
        dataloader = torch.utils.data.DataLoader(X, batch_size=self.batch_size, shuffle=False, num_workers=2)
        predicted_X = None
        for batch_idx, data in enumerate(dataloader):
            if isinstance(data, list) or isinstance(data, tuple):
                inputs = data[0] #For standard datasets from torchvision, data is a list with X and y
            inputs = inputs.to(self.device)
            inputs = Variable(inputs, volatile=True)
            outputs = net(inputs)

            _, predicted = torch.max(outputs.data, 1)
           
            predicted = predicted.detach().cpu().numpy()
            predicted = np.reshape(predicted, (predicted.shape[0], 1))
            if predicted_X is None:
                predicted_X = predicted
            else:
                predicted_X = np.vstack((predicted_X, predicted))
        self.model.train()
        return predicted_X

_input_schema_fit = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Input data schema for training.',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Pytorch Dataset.'
      }}}

_input_predict_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Input data schema for predictions.',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Pytorch Dataset.'
      }}}

_output_predict_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Output data schema for transformed data.',
  'type': 'array',
  'items': {'type': 'array', 'items': {'type': 'number'}}}
  

_hyperparams_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Hyperparameter schema.',
  'allOf': [
    { 'description':
        'This first sub-object lists all constructor arguments with their '
        'types, one at a time, omitting cross-argument constraints.',
      'type': 'object',
      'additionalProperties': False,
      'required': ['num_classes', 'num_epochs',
       'batch_size', 'learning_rate'],
      'relevantToOptimizer': ['num_epochs', 'batch_size', 'learning_rate'],
      'properties': {
        'num_classes': {
          'description': 'Number of classes.',
          'type': 'integer',
          'default': 10,
          'minimum': 2},
        'num_epochs':{
          'description': 'The number of epochs used for training.',
          'type': 'integer',
          'default': 2,
          'minimum': 0,
          'distribution': 'uniform',
          'minimumForOptimizer': 2,
          'maximumForOptimizer': 200},
        'batch_size':{
          'description': 'Batch size used for training and prediction',
          'type': 'integer',
          'default': 64,
          'minimum': 1,
          'distribution': 'uniform',
          'maximumForOptimizer': 128},
        'learning_rate':{
          'description': 'Learning rate scheme for training.',
          'enum': ['constant', 'decay'],
          'default': 'constant',
        },           
        'learning_rate_init':{
          'description': 'Initial value of learning rate for training.',
          'type': 'number',
          'default': 0.1,
          'minimum': 0,
          'distribution': 'loguniform',
          'minimumForOptimizer': 1e-05,
          'maximumForOptimizer': 0.1
        }}}]}

_combined_schemas = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Combined schema for expected data and hyperparameters for a transformer for'
                ' pytorch implementation of ResNet50 for image classification.',
  'type': 'object',
  'tags': {
    'pre': ['images'],
    'op': ['estimator', 'classifier', '~interpretable', 'pytorch'],
    'post': []},
  'properties': {
    'hyperparams': _hyperparams_schema,
    'input_fit': _input_schema_fit,
    'input_predict': _input_predict_schema,
    'output_predict': _output_predict_schema}}

lale.docstrings.set_docstrings(ResNet50Impl, _combined_schemas)

ResNet50 = lale.operators.make_operator(ResNet50Impl, _combined_schemas)

if __name__ == "__main__":
    import torchvision.datasets as datasets

    transform_train = transforms.Compose([
        transforms.RandomCrop(32, padding=4),
        transforms.RandomHorizontalFlip(),
        transforms.ToTensor(),
        transforms.Normalize(mean=[n/255.
                        for n in [129.3, 124.1, 112.4]], std=[n/255. for n in [68.2,  65.4,  70.4]])
    ]) # meanstd transformation

    transform_test = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize(mean=[n/255.
                        for n in [129.3, 124.1, 112.4]], std=[n/255. for n in [68.2,  65.4,  70.4]])
    ])

    data_train = datasets.CIFAR10(root = "/tmp/", download = True, transform = transform_train)
    clf = ResNet50(num_classes=10,num_epochs = 1)
    fitted_clf = clf.fit(data_train)
    data_test = datasets.CIFAR10(root = "/tmp/", download = True, train = False, transform = transform_test)
    predicted = fitted_clf.predict(data_test)
    from sklearn.metrics import accuracy_score

    print(data_test.test_labels[0:10], predicted[0:10])
    print(accuracy_score(data_test.test_labels, predicted.tolist()))



from lale.operators import PlannedOperator, Operator, BasePipeline, OperatorChoice, IndividualOp
from lale.operators import make_choice, make_pipeline, get_pipeline_of_applicable_type
from lale.lib.lale import NoOp
from typing import Optional
from lale.sklearn_compat import clone_op
import random

class NonTerminal(Operator):
    """ Abstract operator for non-terminal grammar rules.
    """
    def __init__(self, name):
        self._name = name
        
    def _lale_clone(self, cloner):
        return NonTerminal(self.name())
    
    def has_same_impl(self):
        pass
    
    def is_supervised(self):
        return False

    def name(self):
        return self._name
    
    def set_name(self, name):
        self._name = name

    def validate_schema(self, X, y=None):
        raise NotImplementedError() #TODO

    def transform_schema(self, s_X):
        raise NotImplementedError() #TODO

    def input_schema_fit(self):
        raise NotImplementedError() #TODO
        
        
class Grammar(Operator):
    """ Base class for Lale grammars.
    """
    def __init__(self):
        self._variables = {}

    def __getattr__(self, name):
        if name.startswith('_'):
            return self.__dict__[name]
        if name not in self._variables:
            self._variables[name] = NonTerminal(name)
        return clone_op(self._variables[name])
        
    def __setattr__(self, name, value):
        if name.startswith('_'):
            self.__dict__[name] = value
        else:
            self._variables[name] = value
            
    def _lale_clone(self):
        pass
    
    def has_same_impl(self):
        pass
    
    def is_supervised(self):
        return False

    def name(self):
        return self._name
    
    def set_name(self, name):
        self._name = name
            
    def validate_schema(self, X, y=None):
        raise NotImplementedError() #TODO

    def transform_schema(self, s_X):
        raise NotImplementedError() #TODO

    def input_schema_fit(self):
        raise NotImplementedError() #TODO

    def _unfold(self, op: Operator, n: int) -> Optional[Operator]:
        """ Unroll all possible operators from the grammar `g` starting from    non-terminal `op` after `n` derivations.
        
        Parameters
        ----------
        op : Operator
            starting rule (e.g., `g.start`)
        n : int
            number of derivations
        
        Returns
        -------
        Optional[Operator]
        """
        if isinstance(op, BasePipeline):
            steps = op.steps()
            new_steps = [self._unfold(sop, n) for sop in op.steps()]
            step_map = {steps[i]: new_steps[i] for i in range(len(steps))}
            new_edges = [(step_map[s], step_map[d]) for s, d in op.edges()]
            if not None in new_steps:
                return get_pipeline_of_applicable_type(new_steps, new_edges, True)
            return None
        if isinstance(op, OperatorChoice):
            steps = [s for s in (self._unfold(sop, n) for sop in op.steps()) if s]
            return make_choice(*steps) if steps else None
        if isinstance(op, NonTerminal):
            return self._unfold(self._variables[op.name()], n-1) if n > 0 else None
        if isinstance(op, IndividualOp):
            return op
        assert False, f"Unknown operator {op}"
                
    def unfold(self, n: int) -> PlannedOperator:
        """
        Explore the grammar `g` starting from `g.start` and generate all possible   choices after `n` derivations.
        
        Parameters
        ----------
        g : Grammar
            input grammar
        n : int
            number of derivations
        
        Returns
        -------
        PlannedOperator
        """
        assert hasattr(self, 'start'), "Rule start must be defined"
        op = self._unfold(self.start, n)
        return make_pipeline(op) if op else NoOp
    
    def _sample(self, op, n):
        """
        Sample the grammar `g` starting from `g.start`, that is, choose one element at random for each possible choices.
        
        Parameters
        ----------
        op : Operator
            starting rule (e.g., `g.start`)
        n : int
            number of derivations
        
        Returns
        -------
        PlannedOperator
        """
        if isinstance(op, BasePipeline):
            steps = op.steps()
            new_steps = [self._sample(sop, n) for sop in op.steps()]
            step_map = {steps[i]: new_steps[i] for i in range(len(steps))}
            new_edges = [(step_map[s], step_map[d]) for s, d in op.edges()]
            if not None in new_steps:
                return get_pipeline_of_applicable_type(new_steps, new_edges, True)
            return None
        if isinstance(op, OperatorChoice):
            return self._sample(random.choice(op.steps()), n)
        if isinstance(op, NonTerminal):
            return self._sample(getattr(self, op.name()), n-1) if n > 0 else None
        if isinstance(op, IndividualOp):
            return op
        assert False, f"Unknown operator {op}"
            
    def sample(self, n: int) -> PlannedOperator:
        """
        Sample the grammar `g` starting from `g.start`, that is, choose one element at random for each possible choices.
          
        Parameters
        ----------
        n : int
            number of derivations
        
        Returns
        -------
        PlannedOperator
        """
        assert hasattr(self, 'start'), "Rule start must be defined"
        op = self._sample(self.start, n)
        return make_pipeline(op) if op else NoOp




from sklearn.gaussian_process.gpr import GaussianProcessRegressor as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class GaussianProcessRegressorImpl():

    def __init__(self, kernel=None, alpha=1e-10, optimizer='fmin_l_bfgs_b', n_restarts_optimizer=0, normalize_y=False, copy_X_train=True, random_state=None):
        self._hyperparams = {
            'kernel': kernel,
            'alpha': alpha,
            'optimizer': optimizer,
            'n_restarts_optimizer': n_restarts_optimizer,
            'normalize_y': normalize_y,
            'copy_X_train': copy_X_train,
            'random_state': random_state}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for GaussianProcessRegressor    Gaussian process regression (GPR).',
    'allOf': [{
        'type': 'object',
        'required': ['kernel', 'alpha', 'optimizer', 'n_restarts_optimizer', 'normalize_y', 'copy_X_train', 'random_state'],
        'relevantToOptimizer': ['alpha', 'optimizer', 'n_restarts_optimizer', 'normalize_y'],
        'additionalProperties': False,
        'properties': {
            'kernel': {
                'XXX TODO XXX': 'kernel object',
                'description': 'The kernel specifying the covariance function of the GP',
                'enum': [None],
                'default': None},
            'alpha': {
                'anyOf': [{
                    'type': 'number',
                    'minimumForOptimizer': 1e-10,
                    'maximumForOptimizer': 1.0,
                    'distribution': 'uniform'}, {
                    'type': 'array',
                    'items': {
                        'laleType': 'Any',
                        'XXX TODO XXX': 'item type'},
                    'XXX TODO XXX': 'float or array-like, optional (default: 1e-10)',
                    'forOptimizer': False}],
                'default': 1e-10,
                'description': 'Value added to the diagonal of the kernel matrix during fitting'},
            'optimizer': {
                'anyOf': [{
                    'type': 'object',
                    'forOptimizer': False}, {
                    'enum': ['fmin_l_bfgs_b']}],
                'default': 'fmin_l_bfgs_b',
                'description': "Can either be one of the internally supported optimizers for optimizing the kernel's parameters, specified by a string, or an externally defined optimizer passed as a callable"},
            'n_restarts_optimizer': {
                'type': 'integer',
                'minimumForOptimizer': 0,
                'maximumForOptimizer': 1,
                'distribution': 'uniform',
                'default': 0,
                'description': "The number of restarts of the optimizer for finding the kernel's parameters which maximize the log-marginal likelihood"},
            'normalize_y': {
                'type': 'boolean',
                'default': False,
                'description': 'Whether the target values y are normalized, i.e., the mean of the observed target values become zero'},
            'copy_X_train': {
                'type': 'boolean',
                'default': True,
                'description': 'If True, a persistent copy of the training data is stored in the object'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The generator used to initialize the centers'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit Gaussian process regression model.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training data'},
        'y': {
            'laleType': 'Any',
            'XXX TODO XXX': 'array-like, shape = (n_samples, [n_output_dims])',
            'description': 'Target values'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict using the Gaussian process regression model',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Query points where the GP is evaluated'},
        'return_std': {
            'type': 'boolean',
            'default': False,
            'description': 'If True, the standard-deviation of the predictive distribution at the query points is returned along with the mean.'},
        'return_cov': {
            'type': 'boolean',
            'default': False,
            'description': 'If True, the covariance of the joint predictive distribution at the query points is returned along with the mean'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict using the Gaussian process regression model',
    'laleType': 'Any',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.gaussian_process.GaussianProcessRegressor#sklearn-gaussian_process-gaussianprocessregressor',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'regressor'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema},
}
lale.docstrings.set_docstrings(GaussianProcessRegressorImpl, _combined_schemas)
GaussianProcessRegressor = lale.operators.make_operator(GaussianProcessRegressorImpl, _combined_schemas)





from sklearn.kernel_approximation import Nystroem as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class NystroemImpl():

    def __init__(self, kernel='rbf', gamma=None, coef0=None, degree=None, kernel_params=None, n_components=100, random_state=None):
        self._hyperparams = {
            'kernel': kernel,
            'gamma': gamma,
            'coef0': coef0,
            'degree': degree,
            'kernel_params': kernel_params,
            'n_components': n_components,
            'random_state': random_state}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for Nystroem    Approximate a kernel map using a subset of the training data.',
    'allOf': [{
        'type': 'object',
        'required': ['kernel', 'gamma', 'coef0', 'degree', 'kernel_params', 'n_components', 'random_state'],
        'relevantToOptimizer': ['kernel', 'n_components'],
        'additionalProperties': False,
        'properties': {
            'kernel': {
                'anyOf': [{
                    'type': 'object',
                    'forOptimizer': False}, {
                    'enum': ['linear', 'poly', 'rbf', 'sigmoid']}],
                'default': 'rbf',
                'description': 'Kernel map to be approximated'},
            'gamma': {
                'anyOf': [{
                    'type': 'number'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Gamma parameter for the RBF, laplacian, polynomial, exponential chi2 and sigmoid kernels'},
            'coef0': {
                'anyOf': [{
                    'type': 'number'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Zero coefficient for polynomial and sigmoid kernels'},
            'degree': {
                'anyOf': [{
                    'type': 'number'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Degree of the polynomial kernel'},
            'kernel_params': {
                'XXX TODO XXX': 'mapping of string to any, optional',
                'description': 'Additional parameters (keyword arguments) for kernel function passed as callable object.',
                'enum': [None],
                'default': None},
            'n_components': {
                'type': 'integer',
                'minimumForOptimizer': 2,
                'maximumForOptimizer': 256,
                'distribution': 'uniform',
                'default': 100,
                'description': 'Number of features to construct'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`.'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit estimator to data.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training data.'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Apply feature map to X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Data to transform.'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transformed data.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.kernel_approximation.Nystroem#sklearn-kernel_approximation-nystroem',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(NystroemImpl, _combined_schemas)
Nystroem = lale.operators.make_operator(NystroemImpl, _combined_schemas)





from sklearn.preprocessing._encoders import OneHotEncoder as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class OneHotEncoderImpl():

    def __init__(self, categories=None, sparse=True, dtype=None, handle_unknown='error', n_values=None, categorical_features=None):
        self._hyperparams = {
            'categories': categories,
            'sparse': sparse,
            'dtype': dtype,
            'handle_unknown': handle_unknown,
            'n_values': n_values,
            'categorical_features': categorical_features}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for OneHotEncoder    Encode categorical integer features as a one-hot numeric array.',
    'allOf': [{
        'type': 'object',
        'required': ['categories', 'sparse', 'dtype', 'handle_unknown', 'n_values', 'categorical_features'],
        'relevantToOptimizer': ['sparse'],
        'additionalProperties': False,
        'properties': {
            'categories': {
                'XXX TODO XXX': "'auto' or a list of lists/arrays of values, default='auto'.",
                'description': "Categories (unique values) per feature:  - 'auto' : Determine categories automatically from the training data",
                'enum': [None],
                'default': None},
            'sparse': {
                'type': 'boolean',
                'default': True,
                'description': 'Will return sparse matrix if set True else will return an array.'},
            'dtype': {
                'laleType': 'Any',
                'XXX TODO XXX': 'number type, default=np.float',
                'description': 'Desired dtype of output.'},
            'handle_unknown': {
                'XXX TODO XXX': "'error' or 'ignore', default='error'.",
                'description': 'Whether to raise an error or ignore if an unknown categorical feature is present during transform (default is to raise)',
                'enum': ['error'],
                'default': 'error'},
            'n_values': {
                'XXX TODO XXX': "'auto', int or array of ints, default='auto'",
                'description': 'Number of values per feature',
                'enum': [None],
                'default': None},
            'categorical_features': {
                'XXX TODO XXX': "'all' or array of indices or mask, default='all'",
                'description': 'Specify what features are treated as categorical',
                'enum': [None],
                'default': None},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit OneHotEncoder to X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The data to determine the categories of each feature.'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform X using one-hot encoding.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The data to encode.'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transformed input.',
    'laleType': 'Any',
    'XXX TODO XXX': 'sparse matrix if sparse=True else a 2-d array',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.OneHotEncoder#sklearn-preprocessing-onehotencoder',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(OneHotEncoderImpl, _combined_schemas)
OneHotEncoder = lale.operators.make_operator(OneHotEncoderImpl, _combined_schemas)





from sklearn.ensemble.forest import ExtraTreesRegressor as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class ExtraTreesRegressorImpl():

    def __init__(self, n_estimators=10, criterion='mse', max_depth=None, min_samples_split=2, min_samples_leaf=1, min_weight_fraction_leaf=0.0, max_features='auto', max_leaf_nodes=None, min_impurity_decrease=0.0, min_impurity_split=None, bootstrap=False, oob_score=False, n_jobs=None, random_state=None, verbose=0, warm_start=False):
        self._hyperparams = {
            'n_estimators': n_estimators,
            'criterion': criterion,
            'max_depth': max_depth,
            'min_samples_split': min_samples_split,
            'min_samples_leaf': min_samples_leaf,
            'min_weight_fraction_leaf': min_weight_fraction_leaf,
            'max_features': max_features,
            'max_leaf_nodes': max_leaf_nodes,
            'min_impurity_decrease': min_impurity_decrease,
            'min_impurity_split': min_impurity_split,
            'bootstrap': bootstrap,
            'oob_score': oob_score,
            'n_jobs': n_jobs,
            'random_state': random_state,
            'verbose': verbose,
            'warm_start': warm_start}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for ExtraTreesRegressor    An extra-trees regressor.',
    'allOf': [{
        'type': 'object',
        'required': ['n_estimators', 'criterion', 'max_depth', 'min_samples_split', 'min_samples_leaf', 'min_weight_fraction_leaf', 'max_features', 'max_leaf_nodes', 'min_impurity_decrease', 'min_impurity_split', 'bootstrap', 'oob_score', 'n_jobs', 'random_state', 'verbose', 'warm_start'],
        'relevantToOptimizer': ['n_estimators', 'criterion', 'max_depth', 'min_samples_split', 'min_samples_leaf', 'max_features', 'bootstrap'],
        'additionalProperties': False,
        'properties': {
            'n_estimators': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 100,
                'distribution': 'uniform',
                'default': 10,
                'description': 'The number of trees in the forest'},
            'criterion': {
                'enum': ['friedman_mse', 'mse'],
                'default': 'mse',
                'description': 'The function to measure the quality of a split'},
            'max_depth': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 3,
                    'maximumForOptimizer': 5,
                    'distribution': 'uniform'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The maximum depth of the tree'},
            'min_samples_split': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 2,
                    'maximumForOptimizer': 5,
                    'distribution': 'uniform'}, {
                    'type': 'number',
                    'forOptimizer': False}],
                'default': 2,
                'description': 'The minimum number of samples required to split an internal node:  - If int, then consider `min_samples_split` as the minimum number'},
            'min_samples_leaf': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 1,
                    'maximumForOptimizer': 5,
                    'distribution': 'uniform'}, {
                    'type': 'number',
                    'forOptimizer': False}],
                'default': 1,
                'description': 'The minimum number of samples required to be at a leaf node'},
            'min_weight_fraction_leaf': {
                'type': 'number',
                'default': 0.0,
                'description': 'The minimum weighted fraction of the sum total of weights (of all the input samples) required to be at a leaf node'},
            'max_features': {
                'anyOf': [{
                    'type': 'integer',
                    'forOptimizer': False}, {
                    'type': 'number',
                    'minimumForOptimizer': 0.01,
                    'maximumForOptimizer': 1.0,
                    'distribution': 'uniform'}, {
                    'type': 'string',
                    'forOptimizer': False}, {
                    'enum': [None]}],
                'default': 'auto',
                'description': 'The number of features to consider when looking for the best split:  - If int, then consider `max_features` features at each split'},
            'max_leaf_nodes': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Grow trees with ``max_leaf_nodes`` in best-first fashion'},
            'min_impurity_decrease': {
                'type': 'number',
                'default': 0.0,
                'description': 'A node will be split if this split induces a decrease of the impurity greater than or equal to this value'},
            'min_impurity_split': {
                'anyOf': [{
                    'type': 'number'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Threshold for early stopping in tree growth'},
            'bootstrap': {
                'type': 'boolean',
                'default': False,
                'description': 'Whether bootstrap samples are used when building trees'},
            'oob_score': {
                'type': 'boolean',
                'default': False,
                'description': 'Whether to use out-of-bag samples to estimate the R^2 on unseen data.'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The number of jobs to run in parallel for both `fit` and `predict`'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`.'},
            'verbose': {
                'type': 'integer',
                'default': 0,
                'description': 'Controls the verbosity when fitting and predicting.'},
            'warm_start': {
                'type': 'boolean',
                'default': False,
                'description': 'When set to ``True``, reuse the solution of the previous call to fit and add more estimators to the ensemble, otherwise, just fit a whole new forest'},
        }}, {
        'XXX TODO XXX': 'Parameter: min_samples_leaf > only be considered if it leaves at least min_samples_leaf training samples in each of the left and right branches'}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Build a forest of trees from the training set (X, y).',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array-like or sparse matrix of shape = [n_samples, n_features]'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'The training input samples'},
        'y': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'The target values (class labels in classification, real numbers in regression).'},
        'sample_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'enum': [None]}],
            'description': 'Sample weights'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict regression target for X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array-like or sparse matrix of shape = [n_samples, n_features]'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'The input samples'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'The predicted values.',
    'anyOf': [{
        'type': 'array',
        'items': {
            'type': 'number'},
    }, {
        'type': 'array',
        'items': {
            'type': 'array',
            'items': {
                'type': 'number'},
        }}],
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.ensemble.ExtraTreesRegressor#sklearn-ensemble-extratreesregressor',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'regressor'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema},
}
lale.docstrings.set_docstrings(ExtraTreesRegressorImpl, _combined_schemas)
ExtraTreesRegressor = lale.operators.make_operator(ExtraTreesRegressorImpl, _combined_schemas)





from sklearn.cross_decomposition.pls_ import PLSCanonical as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class PLSCanonicalImpl():

    def __init__(self, n_components=2, scale=True, algorithm='nipals', max_iter=500, tol=1e-06, copy=True):
        self._hyperparams = {
            'n_components': n_components,
            'scale': scale,
            'algorithm': algorithm,
            'max_iter': max_iter,
            'tol': tol,
            'copy': copy}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)

    def predict(self, X):
        return self._wrapped_model.predict(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for PLSCanonical    PLSCanonical implements the 2 blocks canonical PLS of the original Wold',
    'allOf': [{
        'type': 'object',
        'required': ['n_components', 'scale', 'algorithm', 'max_iter', 'tol', 'copy'],
        'relevantToOptimizer': ['n_components', 'scale', 'algorithm', 'max_iter', 'tol', 'copy'],
        'additionalProperties': False,
        'properties': {
            'n_components': {
                'type': 'integer',
                'minimumForOptimizer': 2,
                'maximumForOptimizer': 256,
                'distribution': 'uniform',
                'default': 2,
                'description': 'Number of components to keep'},
            'scale': {
                'type': 'boolean',
                'default': True,
                'description': 'Option to scale data'},
            'algorithm': {
                'XXX TODO XXX': 'string, "nipals" or "svd"',
                'description': 'The algorithm used to estimate the weights',
                'enum': ['nipals', 'svd'],
                'default': 'nipals'},
            'max_iter': {
                'XXX TODO XXX': 'an integer, (default 500)',
                'description': 'the maximum number of iterations of the NIPALS inner loop (used only if algorithm="nipals")',
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 500},
            'tol': {
                'XXX TODO XXX': 'non-negative real, default 1e-06',
                'description': 'the tolerance used in the iterative algorithm',
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'uniform',
                'default': 1e-06},
            'copy': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether the deflation should be done on a copy'},
        }}, {
        'XXX TODO XXX': 'Parameter: max_iter > only if algorithm="nipals")'}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit model to data.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vectors, where n_samples is the number of samples and n_features is the number of predictors.'},
        'Y': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Target vectors, where n_samples is the number of samples and n_targets is the number of response variables.'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Apply the dimension reduction learned on the train data.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vectors, where n_samples is the number of samples and n_features is the number of predictors.'},
        'Y': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Target vectors, where n_samples is the number of samples and n_targets is the number of response variables.'},
        'copy': {
            'type': 'boolean',
            'default': True,
            'description': 'Whether to copy X and Y, or perform in-place normalization.'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Apply the dimension reduction learned on the train data.',
    'laleType': 'Any',
    'XXX TODO XXX': 'x_scores if Y is not given, (x_scores, y_scores) otherwise.',
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Apply the dimension reduction learned on the train data.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vectors, where n_samples is the number of samples and n_features is the number of predictors.'},
        'copy': {
            'type': 'boolean',
            'default': True,
            'description': 'Whether to copy X and Y, or perform in-place normalization.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Apply the dimension reduction learned on the train data.',
    'laleType': 'Any',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.cross_decomposition.PLSCanonical#sklearn-cross_decomposition-plscanonical',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer', 'estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema},
}
lale.docstrings.set_docstrings(PLSCanonicalImpl, _combined_schemas)
PLSCanonical = lale.operators.make_operator(PLSCanonicalImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.naive_bayes
import lale.docstrings
import lale.operators

class MultinomialNBImpl():

    def __init__(self, alpha=1.0, fit_prior=True, class_prior=None):
        self._hyperparams = {
            'alpha': alpha,
            'fit_prior': fit_prior,
            'class_prior': class_prior}
        self._wrapped_model = sklearn.naive_bayes.MultinomialNB(
            **self._hyperparams)

    def fit(self, X, y=None):
        self._wrapped_model.fit(X, y)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)


_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Naive Bayes classifier for multinomial models',
    'allOf': [{
        'type': 'object',
        'required': ['alpha', 'fit_prior'],
        'relevantToOptimizer': ['alpha', 'fit_prior'],       
        'properties': {
            'alpha': {
                'type': 'number',
                'distribution':'loguniform',
                'minimumForOptimizer': 1e-10,
                'maximumForOptimizer': 1.0,
                'default': 1.0,
                'description': 'Additive (Laplace/Lidstone) smoothing parameter'},
            'fit_prior': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether to learn class prior probabilities or not.'},
            'class_prior': {
                'anyOf': [{
                    'type': 'array',
                    'items': {'type': 'number'}}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Prior probabilities of the classes. If specified the priors are not'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit Naive Bayes classifier according to X, y',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {'type': 'array', 'items': {'type': 'number'}},
            'description': 'Training vectors, where n_samples is the number of samples and n_features is the number of features.'},
        'y': {
            'anyOf': [
                {'type': 'array', 'items': {'type': 'number'}},
                {'type': 'array', 'items': {'type': 'string'}}],
            'description': 'Target values.'},
        'sample_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {'type': 'number'}}, {
                'enum': [None]}],
            'default': None,
            'description': 'Weights applied to individual samples (1. for unweighted).'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Perform classification on an array of test vectors X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {'type': 'array', 'items': {'type': 'number'}}},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Perform classification on an array of test vectors X.',
    'anyOf': [
        {'type': 'array', 'items': {'type': 'number'}},
        {'type': 'array', 'items': {'type': 'string'}}]}

_input_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Perform classification on an array of test vectors X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {'type': 'array', 'items': {'type': 'number'}}},
    },
}
_output_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Perform classification on an array of test vectors X.',
    'type': 'array',
    'items': {'type': 'array', 'items': {'type': 'number'}},
    'description': 'Returns the probability of the samples for each class in the model. The columns correspond to the classes in sorted order, as they appear in the attribute classes_.'
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.naive_bayes.MultinomialNB.html#sklearn-naive-bayes-multinomialnb',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'classifier'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_predict_proba': _input_predict_proba_schema,
        'output_predict_proba': _output_predict_proba_schema}}

lale.docstrings.set_docstrings(MultinomialNBImpl, _combined_schemas)

MultinomialNB = lale.operators.make_operator(MultinomialNBImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lale uses `JSON Schema`_ to check machine-learning pipelines for correct types.

In general, there are two kinds of checks. The first is an instance
check (`v: s`), which checks whether a JSON value v is valid for a
schema s. The second is a subschema_ check (`s <: t`), which checks
whether one schema s is a subchema of another schema t.

Besides regular JSON values, Lale also supports certain JSON-like
values. For example, a ``np.ndarray`` of numbers is treated like a
JSON array of arrays of numbers. Furthermore, Lale supports an 'Any'
type for which all instance and subschema checks on the left as well
as the right side succeed. This is specified using ``{'laleType': 'Any'}``.

.. _`JSON Schema`: https://json-schema.org/understanding-json-schema/reference/

.. _subschema: https://arxiv.org/abs/1911.12651
"""

import functools
import jsonschema
import jsonsubschema
import lale.helpers
import numpy as np
import pandas as pd
import scipy.sparse
import logging
import inspect
from typing import Any, Dict, List, Union
JSON_TYPE = Dict[str, Any]

def validate_schema(value, schema: JSON_TYPE, subsample_array:bool=True):
    """Validate that the value is an instance of the schema.

    Parameters
    ----------
    value: JSON (int, float, str, list, dict) or JSON-like (tuple, np.ndarray, pd.DataFrame ...).
        Left-hand side of instance check.

    schema: JSON schema
        Right-hand side of instance check.

    subsample_array: bool
        Speed up checking by doing only partial conversion to JSON.

    Raises
    ------
    jsonschema.ValidationError
        The value was invalid for the schema.
    """
    json_value = lale.helpers.data_to_json(value, subsample_array)
    jsonschema.validate(json_value, schema, jsonschema.Draft4Validator)

_JSON_META_SCHEMA_URL = 'http://json-schema.org/draft-04/schema#'

def _json_meta_schema() -> Dict[str, Any]:
    return jsonschema.Draft4Validator.META_SCHEMA

def validate_is_schema(value: Dict[str, Any]):
    #TODO: move this function to lale.type_checking
    if '$schema' in value:
        assert value['$schema'] == _JSON_META_SCHEMA_URL
    jsonschema.validate(value, _json_meta_schema())

def is_schema(value) -> bool:
    if isinstance(value, dict):
        try:
            jsonschema.validate(value, _json_meta_schema())
        except:
            return False
        return True
    return False

def _json_replace(subject, old, new):
    if subject == old:
        return new
    if isinstance(subject, list):
        result = [_json_replace(s, old, new) for s in subject]
        for i in range(len(subject)):
            if subject[i] != result[i]:
                return result
    elif isinstance(subject, tuple):
        result = tuple([_json_replace(s, old, new) for s in subject])
        for i in range(len(subject)):
            if subject[i] != result[i]:
                return result
    elif isinstance(subject, dict):
        if isinstance(old, dict):
            is_sub_dict = True
            for k, v in old.items():
                if k not in subject or subject[k] != v:
                    is_sub_dict = False
                    break
            if is_sub_dict:
                return new
        result = {k: _json_replace(v, old, new) for k, v in subject.items()}
        for k in subject:
            if subject[k] != result[k]:
                return result
    return subject #nothing changed so share original object (not a copy)

def is_subschema(sub_schema, super_schema) -> bool:
    """Is sub_schema a subschema of super_schema?

    Parameters
    ----------
    sub_schema: JSON schema
        Left-hand side of subschema check.

    super_schema: JSON schema
        Right-hand side of subschema check.

    Returns
    -------
    bool
        True if `sub_schema <: super_schema`, False otherwise.
    """
    new_sub = _json_replace(sub_schema, {'laleType': 'Any'}, {'not': {}})
    try:
        return jsonsubschema.isSubschema(new_sub, super_schema)
    except Exception as e:
        raise ValueError(f'unexpected internal error checking ({new_sub} <: {super_schema})') from e

class SubschemaError(Exception):
    """Raised when a subschema check (sub `<:` sup) failed.
    """
    def __init__(self, sub, sup, sub_name='sub', sup_name='super'):
        self.sub = sub
        self.sup = sup
        self.sub_name = sub_name
        self.sup_name = sup_name

    def __str__(self):
        summary = f'Expected {self.sub_name} to be a subschema of {self.sup_name}.'
        import lale.pretty_print
        sub = lale.pretty_print.schema_to_string(self.sub)
        sup = lale.pretty_print.schema_to_string(self.sup)
        details = f'\n{self.sub_name} = {sub}\n{self.sup_name} = {sup}'
        return summary + details

def _validate_subschema(sub, sup, sub_name='sub', sup_name='super'):
    if not is_subschema(sub, sup):
        raise SubschemaError(sub, sup, sub_name, sup_name)

def validate_schema_or_subschema(lhs, super_schema):
    """Validate that lhs is an instance of or a subschema of super_schema.

    Parameters
    ----------
    lhs: value or JSON schema
        Left-hand side of instance or subschema check.

    super_schema: JSON schema
        Right-hand side of instance or subschema check.

    Raises
    ------
    jsonschema.ValidationError
        The lhs was an invalid value for super_schema.

    SubschemaError
        The lhs was or had a schema that was not a subschema of super_schema.
    """
    if is_schema(lhs):
        sub_schema = lhs
    else:
        import lale.datasets.data_schemas
        try:
            sub_schema = lale.datasets.data_schemas.to_schema(lhs)
        except ValueError as e:
            sub_schema = None
    if sub_schema is None:
        validate_schema(lhs, super_schema)
    else:
        _validate_subschema(sub_schema, super_schema)

def join_schemas(*schemas: JSON_TYPE) -> JSON_TYPE:
    """Compute the lattice join (union type, disjunction) of the arguments.

    Parameters
    ----------
    *schemas: list of JSON schemas
        Schemas to be joined.

    Returns
    -------
    JSON schema
        The joined schema.
    """
    def join_two_schemas(s_a, s_b):
        if s_a is None:
            return s_b
        s_a = lale.helpers.dict_without(s_a, 'description')
        s_b = lale.helpers.dict_without(s_b, 'description')
        if is_subschema(s_a, s_b):
            return s_b
        if is_subschema(s_b, s_a):
            return s_a
        return jsonsubschema.joinSchemas(s_a, s_b)
    if len(schemas) == 0:
        return {'not':{}}
    result = functools.reduce(join_two_schemas, schemas)
    return result
        
def get_hyperparam_names(op: 'lale.operators.IndividualOp') -> List[str]:
    """Names of the arguments to the constructor of the impl.

    Parameters
    ----------
    op: lale.operators.IndividualOp
        Operator whose hyperparameters to get.

    Returns
    -------
    List[str]
        List of hyperparameter names.
    """
    if op._impl.__module__.startswith('lale'):
        hp_schema = op.hyperparam_schema()
        params = next(iter(hp_schema.get('allOf', []))).get('properties', {})
        return list(params.keys())
    else:
        return inspect.getargspec(op._impl_class().__init__).args
                
def validate_method(op: 'lale.operators.IndividualOp', schema_name: str):
    """Check whether the operator has the given method schema.

    Parameters
    ----------
    op: lale.operators.IndividualOp
        Operator whose methods to check.

    schema_name: 'input_fit' or 'input_predict' or 'input_predict_proba' or 'input_transform' 'output_predict' or 'output_predict_proba' or 'output_transform'
        Name of schema to check.

    Raises
    ------
    AssertionError
        The operator does not have the given schema.
    """
    if op._impl.__module__.startswith('lale'):
        assert schema_name in op._schemas['properties']
    else:
        method_name = ''
        if schema_name.startswith('input_'):
            method_name = schema_name[len('input_'):]
        elif schema_name.startswith('output_'):
            method_name = schema_name[len('output_'):]
        if method_name:
            assert hasattr(op._impl, method_name)

def _get_args_schema(fun):
    sig = inspect.signature(fun)
    result = {'type': 'object', 'properties': {}}
    required = []
    additional_properties = False
    for name, param in sig.parameters.items():
        ignored_kinds = [inspect.Parameter.VAR_POSITIONAL,
                         inspect.Parameter.VAR_KEYWORD]
        if name != 'self':
            if param.kind in ignored_kinds:
                additional_properties = True
            else:
                if param.default == inspect.Parameter.empty:
                    param_schema = {'laleType': 'Any'}
                    required.append(name)
                else:
                    param_schema = {'default': param.default}
                result['properties'][name] = param_schema
    if not additional_properties:
        result['additionalProperties'] = False
    if len(required) > 0:
        result['required'] = required
    return result

def get_default_schema(impl):
    """Creates combined schemas for a bare operator implementation class.

    Used when there were no explicit combined schemas provided when
    the operator was created. The default schema provides defaults by
    inspecting the signature of the ``__init__`` method, and uses
    'Any' types for the inputs and outputs of other methods.

    Returns
    -------
    JSON Schema
        Combined schema with properties for hyperparams and
        all applicable method inputs and outputs.
    """
    if hasattr(impl, '__init__'):
        hyperparams_schema = _get_args_schema(impl.__init__)
    else:
        hyperparams_schema = {'type': 'object', 'properties': {}}
    hyperparams_schema['relevantToOptimizer'] = []
    method_schemas = {'hyperparams': {'allOf': [hyperparams_schema]}}
    if hasattr(impl, 'fit'):
        method_schemas['input_fit'] = _get_args_schema(impl.fit)
    for method_name in ['predict', 'predict_proba', 'transform']:
        if hasattr(impl, method_name):
            method_args_schema = _get_args_schema(getattr(impl, method_name))
            method_schemas['input_' + method_name] = method_args_schema
            method_schemas['output_' + method_name] = {'laleType': 'Any'}
    tags = {
        'pre': [],
        'op': (['transformer'] if hasattr(impl, 'transform') else [])
            + (['estimator'] if hasattr(impl, 'predict') else []),
        'post': []}
    result = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'description': f'Schema for {type(impl)} auto-generated by lale.type_checking.get_default_schema().',
        'type': 'object',
        'tags': tags,
        'properties': method_schemas}
    return result



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from typing import Any, Dict

from sklearn.base import clone

from lale.operators import make_operator
from lale.sklearn_compat import make_sklearn_compat
from lale.lib.lale import ConcatFeatures as Concat


class MutatingOpImpl():
    fit_counter:int
    predict_counter:int
    def __init__(self, k=0):
      self.fit_counter = 0
      self.predict_counter = 0
      self.k = k

    def fit(self, X, y = None):
        assert self.fit_counter == 0
        self.fit_counter = self.fit_counter + 1
        return self

    def predict(self, X, y = None):
        assert self.predict_counter == 0
        self.predict_counter = self.predict_counter + 1
        return [[1] for x in X]

    def get_params(self, deep:bool = True)->Dict[str,Any]:
        out:Dict[str,Any] = {}
        out['k'] = self.k
        return out

    def set_params(self, **impl_params):
        self.k = impl_params['k']
        return self



    # def transform(self, X, y = None):
    #     return X, y

_input_schema_fit = {
  '$schema': 'http://json-schema.org/draft-04/schema#'
}
_input_predict_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
}
_output_predict_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#'}

_hyperparam_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'allOf': [
    { 'description':
        'This first sub-object lists all constructor arguments with their '
        'types, one at a time, omitting cross-argument constraints.',
      'type': 'object',
      'additionalProperties': False,
      'relevantToOptimizer': [],
      'properties': {
        'k': {
          'type': 'number'
        }
      }
    }]}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparam_schema,
        'input_fit': _input_schema_fit,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema}}

MutatingOp = make_operator(MutatingOpImpl, _combined_schemas)

def fit_clone_fit(op):
  op1 = make_sklearn_compat(op)
  op1.fit(X=[1,2], y=[1,2])
  op2 = clone(op1)
  fit2 = op2.fit(X=[3,4], y=[3,4])
  print(fit2)

class TestClone(unittest.TestCase):
  def test_clone_clones_op(self):
    op = MutatingOp(k=1)
    fit_clone_fit(op)
  
  def test_clone_clones_seq(self):
    op = MutatingOp(k=1) >> MutatingOp(k=2)
    fit_clone_fit(op)
  
  def test_clone_clones_and(self):
    op = MutatingOp(k=1) & MutatingOp(k=2)
    fit_clone_fit(op)

  def test_clone_clones_concat(self):
    from lale.lib.lale import NoOp
    op = ((MutatingOp(k=1) & MutatingOp(k=2))) >> Concat | MutatingOp(k=4)

  def test_clone_clones_choice(self):
    op = MutatingOp(k=1) | MutatingOp(k=2)
    fit_clone_fit(op)

  def test_clone_clones_complex(self):
    op = ((MutatingOp(k=1) | ((MutatingOp(k=2) & MutatingOp(k=3)) >> Concat)) >> MutatingOp(k=4)) | MutatingOp(k=5)
    fit_clone_fit(op)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.docstrings
import lale.operators
import sklearn.decomposition

class PCAImpl():
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = sklearn.decomposition.PCA(**self._hyperparams)

    def fit(self, X, y=None):
        self._wrapped_model.fit(X, y)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Hyperparameter schema for the PCA model from scikit-learn.',
    'allOf': [{
        'description': 'This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.\n',
        'type': 'object',
        'additionalProperties': False,
        'required': ['n_components', 'copy', 'whiten', 'svd_solver', 'tol', 'iterated_power', 'random_state'],
        'relevantToOptimizer': ['n_components', 'whiten', 'svd_solver'],
        'properties': {
            'n_components': {
                'anyOf': [{
                    'description': 'If not set, keep all components.',
                    'enum': [None]},
                {   'description': "Use Minka's MLE to guess the dimension.",
                    'enum': ['mle']},
                {   'description': """Select the number of components such that the amount of variance that
needs to be explained is greater than the specified percentage.""",
                    'type': 'number',
                    'minimum': 0.0,
                    'exclusiveMinimum': True,
                    'maximum': 1.0,
                    'exclusiveMaximum': True},
                {   'description': 'Number of components to keep.',
                    'type': 'integer',
                    'minimum': 1,
                    'forOptimizer': False}],
                'default': None},
            'copy': {
                'description': 'If false, overwrite data passed to fit.',
                'type': 'boolean',
                'default': True},
            'whiten': {
                'description': """When true, multiply the components vectors by the square root of
n_samples and then divide by the singular values to ensure uncorrelated
outputs with unit component-wise variances.""",
                'type': 'boolean',
                'default': False},
            'svd_solver': {
                'description': 'Algorithm to use.',
                'enum': ['auto', 'full', 'arpack', 'randomized'],
                'default': 'auto'},
            'tol': {
                'description': 'Tolerance for singular values computed by svd_solver arpack.',
                'type': 'number',
                'minimum': 0.0,
                'default': 0.0},
            'iterated_power': {
                'anyOf': [
                {   'description': 'Number of iterations for the power method computed by svd_solver randomized.',
                    'type': 'integer',
                    'minimum': 0},
                {   'description': 'Pick automatically.',
                    'enum': ['auto']}],
                'default': 'auto'},
            'random_state': {
                'description': 'Seed of pseudo-random number generator for shuffling data.',
                'anyOf': [{
                    'description': 'RandomState used by np.random',
                    'enum': [None]}, {
                    'description': 'Explicit seed.',
                    'type': 'integer'}],
                'default': None},
        }},
    {   'description': 'Option n_components mle can only be set for svd_solver full or auto.',
        'anyOf': [
        {   'type': 'object',
            'properties': {
                'n_components': {
                    'not': {
                        'enum': ['mle']},
                }},
        },
        {   'type': 'object',
            'properties': {
                'svd_solver': {
                    'enum': ['full', 'auto']},
            }}]},
    {   'description': 'Setting 0 < n_components < 1 only works for svd_solver full.',
        'anyOf': [
        {   'type': 'object',
            'properties': {
                'n_components': {
                    'not': {
                        'type': 'number',
                        'minimum': 0.0,
                        'exclusiveMinimum': True,
                        'maximum': 1.0,
                        'exclusiveMaximum': True},
                }}},
        {   'type': 'object',
            'properties': {
                'svd_solver': {
                    'enum': ['full']},
            }}]},
    {   'description': 'Option tol can be set for svd_solver arpack.',
        'anyOf': [
        {   'type': 'object',
            'properties': {
                'tol': {
                    'enum': [0.0]}}},
        {   'type': 'object',
            'properties': {
                'svd_solver': {
                    'enum': ['arpack']},
            }}]},
    {   'description': 'Option iterated_power can be set for svd_solver randomized.',
        'anyOf': [
        {   'type': 'object',
            'properties': {
                'iterated_power': {
                    'enum': ['auto']},
            }},
        {   'type': 'object',
            'properties': {
                'svd_solver': {
                    'enum': ['randomized']},
            }}]},
]}

_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Features; the outer array is over samples.',
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {'type': 'number'}}},
        'y': {
            'description': 'Target for supervised learning (ignored).',
            'laleType': 'Any'}}}

_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Features; the outer array is over samples.',
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {'type': 'number'}}}}}

_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Features; the outer array is over samples.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {'type': 'number'}}}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """`Principal component analysis`_ for linear dimensionality reduction.

.. _`Principal component analysis`: https://scikit-learn.org/0.20/modules/generated/sklearn.decomposition.PCA.html#sklearn-decomposition-pca
""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.pca.html',
    'type': 'object',
    'tags': {
        'pre': ['~categoricals'],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(PCAImpl, _combined_schemas)

PCA = lale.operators.make_operator(PCAImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Iterable, List, Optional, Tuple, Union
import itertools
import warnings
import random
import math
from collections import ChainMap

from lale.util.Visitor import Visitor, accept
from lale.search.search_space import SearchSpace, SearchSpaceObject, SearchSpaceConstant, SearchSpaceEnum, SearchSpaceSum, SearchSpaceProduct, SearchSpacePrimitive, SearchSpaceArray, SearchSpaceOperator, should_print_search_space
from lale.search.schema2search_space import op_to_search_space
from lale.search.PGO import PGO
from lale.sklearn_compat import nest_all_HPparams, nest_choice_all_HPparams, DUMMY_SEARCH_SPACE_GRID_PARAM_NAME, discriminant_name, make_indexed_name, make_array_index_name, structure_type_name, structure_type_list, structure_type_tuple, structure_type_dict

from lale.operators import PlannedOperator, OperatorChoice, PlannedIndividualOp, PlannedPipeline, Operator

SearchSpaceGrid = Dict[str,SearchSpacePrimitive]

def search_space_grid_to_string(grid:SearchSpaceGrid)->str:
    return "{" + ";".join(f"{k}->{str(v)}" for k,v in grid.items()) + "}"

def search_space_grids_to_string(grids:List[SearchSpaceGrid])->str:
    return "|".join(search_space_grid_to_string(grid) for grid in grids)

def get_search_space_grids( op:'PlannedOperator', 
                            num_grids:Optional[float]=None, 
                            pgo:Optional[PGO]=None)->List[SearchSpaceGrid]:
    """ Top level function: given a lale operator, returns a list of hp grids.
    Parameters
    ----------
    op : The lale PlannedOperator
    num_grids: integer or float, optional
        if set to an integer => 1, it will determine how many parameter grids will be returned (at most)
        if set to an float between 0 and 1, it will determine what fraction should be returned
        note that setting it to 1 is treated as in integer.  To return all results, use None
    """
    all_parameters = op_to_search_space_grids(op, pgo=pgo)
    if should_print_search_space("true", "all", "search_space_grids", "grids"):
        name = op.name()
        if not name:
            name = "an operator"
        print(f"search space grids for {name}:\n{search_space_grids_to_string(all_parameters)}")
    if num_grids is None:
        return all_parameters
    else:
        if num_grids <= 0:
            warnings.warn(f"get_search_space_grids(num_grids={num_grids}) called with a non-positive value for lale_num_grids")
            return []
        if num_grids >= 1:
            samples = math.ceil(num_grids)
            if samples >= len(all_parameters):
                return all_parameters
            else:
                warnings.warn(f"get_search_space_grids(num_grids={num_grids}) sampling {math.ceil(num_grids)}/{len(all_parameters)}")
                return random.sample(all_parameters, math.ceil(num_grids))
        else:
            samples = round(len(all_parameters)*num_grids)
            warnings.warn(f"get_search_space_grids(num_grids={num_grids}) sampling {samples}/{len(all_parameters)}")
            return random.sample(all_parameters, samples)

def search_space_to_grids(hp:SearchSpace)->List[SearchSpaceGrid]:
    return SearchSpaceToGridVisitor.run(hp)

def op_to_search_space_grids(op:PlannedOperator,
                             pgo:Optional[PGO]=None)->List[SearchSpaceGrid]:
    search_space = op_to_search_space(op, pgo=pgo)
    grids = search_space_to_grids(search_space)
    return grids

# lets handle the general case
SearchSpaceGridInternalType = Union[List[SearchSpaceGrid], SearchSpacePrimitive]

class SearchSpaceToGridVisitor(Visitor):
    @classmethod
    def run(cls, space:SearchSpace)->List[SearchSpaceGrid]:
        visitor = cls()
        space_:Any = space
        grids:SearchSpaceGridInternalType = space_.accept(visitor)
        fixed_grids = cls.fixupDegenerateSearchSpaces(grids)
        return fixed_grids

    @classmethod
    def fixupDegenerateSearchSpaces(cls, space:SearchSpaceGridInternalType)->List[SearchSpaceGrid]:
        if isinstance(space, SearchSpacePrimitive):
            return [{DUMMY_SEARCH_SPACE_GRID_PARAM_NAME:space}]
        else:
            return space

    def __init__(self):
        super(SearchSpaceToGridVisitor, self).__init__()

    def visitSearchSpacePrimitive(self, space:SearchSpacePrimitive)->SearchSpacePrimitive:
        return space
    
    visitSearchSpaceEnum = visitSearchSpacePrimitive

    visitSearchSpaceConstant = visitSearchSpaceEnum
    visitSearchSpaceBool = visitSearchSpaceEnum

    visitSearchSpaceNumber = visitSearchSpacePrimitive

    def _searchSpaceList(self, space:SearchSpaceArray, *, size:int)->List[SearchSpaceGrid]:
        sub_spaces = space.items(max=size)

        param_grids:List[List[SearchSpaceGrid]] = \
            [nest_all_HPparams(str(index), self.fixupDegenerateSearchSpaces(accept(sub,self))) for index,sub in enumerate(sub_spaces)]

        param_grids_product:Iterable[Iterable[SearchSpaceGrid]] = itertools.product(*param_grids)
        chained_grids:List[SearchSpaceGrid] = [
            dict(ChainMap(*gridline,)) for gridline in param_grids_product]

        if space.is_tuple:
            st_val = structure_type_tuple
        else:
            st_val = structure_type_list

        discriminated_grids:List[SearchSpaceGrid]=[{**d, structure_type_name:SearchSpaceConstant(st_val)} for d in chained_grids]


        return discriminated_grids

    def visitSearchSpaceArray(self, space:SearchSpaceArray)->List[SearchSpaceGrid]:
        if space.minimum == space.maximum:
            return self._searchSpaceList(space, size=space.minimum)
        else:
            ret:List[SearchSpaceGrid] = []
            for i in range(space.minimum, space.maximum+1):
                ret.extend(self._searchSpaceList(space, size=i))
            return ret

    def visitSearchSpaceObject(self, space:SearchSpaceObject)->List[SearchSpaceGrid]:
        keys = space.keys
        keys_len = len(keys)
        final_choices:List[SearchSpaceGrid] = []
        for c in space.choices:
            assert keys_len == len(c)
            kvs_complex:List[List[SearchSpaceGrid]] = []
            kvs_simple:SearchSpaceGrid = {}
            for k,v in zip(keys, c):
                vspace:Union[List[SearchSpaceGrid],SearchSpacePrimitive] = v.accept(self)
                if isinstance(vspace, SearchSpacePrimitive):
                    kvs_simple[k] = vspace
                else:
                    nested_vspace:List[SearchSpaceGrid] = nest_all_HPparams(k, vspace)
                    if nested_vspace:
                        kvs_complex.append(nested_vspace)
            nested_space_choices:Iterable[Iterable[SearchSpaceGrid]] = itertools.product(*kvs_complex)
            nested_space_choices_lists:List[List[SearchSpaceGrid]] = list(map((lambda x: list(x)),nested_space_choices))
            nested_space_choices_filtered:List[List[SearchSpaceGrid]] = [
                l for l in nested_space_choices_lists if l]
            if nested_space_choices_filtered:
                chained_grids:Iterable[SearchSpaceGrid] = [
                  dict(ChainMap(*nested_choice, kvs_simple)) for nested_choice in nested_space_choices_filtered]
                final_choices.extend(chained_grids)
            else:
                final_choices.append(kvs_simple)
        return final_choices

    def visitSearchSpaceSum(self, op:SearchSpaceSum)->SearchSpaceGridInternalType:
        sub_spaces:List[SearchSpace] = op.sub_spaces

        sub_grids:Iterable[SearchSpaceGridInternalType] = [
            accept(cur_space, self) for cur_space in sub_spaces]

        if len(sub_spaces) == 1:
            return list(sub_grids)[0]
        else:
            fixed_grids:Iterable[List[SearchSpaceGrid]] = (SearchSpaceToGridVisitor.fixupDegenerateSearchSpaces(grid) for grid in sub_grids)
            final_grids:List[SearchSpaceGrid] = []
            for i, grids in enumerate(fixed_grids):
                if not grids:
                    grids = [{}]
                else:
                    # we need to add in this nesting
                    # in case a higher order operator directly contains
                    # another
                    grids = nest_choice_all_HPparams(grids)

                discriminated_grids:List[SearchSpaceGrid]=[{**d, discriminant_name:SearchSpaceConstant(i)} for d in grids]
                final_grids.extend(discriminated_grids)
            return final_grids

    def visitSearchSpaceProduct(self, op:SearchSpaceProduct)->SearchSpaceGridInternalType:

        sub_spaces = op.get_indexed_spaces()

        param_grids:List[List[SearchSpaceGrid]] = [nest_all_HPparams(make_indexed_name(name, index), self.fixupDegenerateSearchSpaces(accept(space,self))) for name,index,space in sub_spaces]

        param_grids_product:Iterable[Iterable[SearchSpaceGrid]] = itertools.product(*param_grids)
        chained_grids:List[SearchSpaceGrid] = [
            dict(ChainMap(*gridline)) for gridline in param_grids_product]

        return chained_grids

    def visitSearchSpaceOperator(self, op:SearchSpaceOperator)->SearchSpaceGridInternalType:
        return accept(op.sub_space, self)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.linear_model.passive_aggressive
import lale.docstrings
import lale.operators

class PassiveAggressiveClassifierImpl():

    def __init__(self, C=1.0, fit_intercept=True, max_iter=None, tol=None, early_stopping=False, 
    validation_fraction=0.1, n_iter_no_change=5, shuffle=True, verbose=0, loss='hinge', 
    n_jobs=None, random_state=None, warm_start=False, class_weight=None, average=False):
    #The wrapper does not support n_iter as it is deprecated and will be removed in sklearn 0.21.
        self._hyperparams = {
            'C': C,
            'fit_intercept': fit_intercept,
            'max_iter': max_iter,
            'tol': tol,
            'early_stopping': early_stopping,
            'validation_fraction': validation_fraction,
            'n_iter_no_change': n_iter_no_change,
            'shuffle': shuffle,
            'verbose': verbose,
            'loss': loss,
            'n_jobs': n_jobs,
            'random_state': random_state,
            'warm_start': warm_start,
            'class_weight': class_weight,
            'average': average}
        self._wrapped_model = sklearn.linear_model.passive_aggressive.PassiveAggressiveClassifier(**self._hyperparams)

    def fit(self, X, y=None):
        self._wrapped_model.fit(X, y)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def decision_function(self, X):
        return self._wrapped_model.decision_function(X)

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Passive Aggressive Classifier',
    'allOf': [{
        'type': 'object',
        'additionalProperties': False,
        'required': ['C', 'fit_intercept', 'max_iter', 'tol', 'early_stopping', 
            'shuffle', 'loss', 'average'],
        'relevantToOptimizer': ['C', 'fit_intercept', 'max_iter', 'tol', 'early_stopping', 
            'shuffle', 'loss', 'average'],         
        'properties': {
            'C': {
                'type': 'number',
                'description': 'Maximum step size (regularization). Defaults to 1.0.',
                'default': 1.0,
                'distribution': 'loguniform',
                'minimumForOptimizer': 1e-5,
                'maximumForOptimizer': 10},
            'fit_intercept': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether the intercept should be estimated or not. If False, the'
                'the data is assumed to be already centered.'},
            'max_iter': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 5,
                    'maximumForOptimizer': 1000,
                    'distribution': 'uniform',
                    'default': 5}, #default value is 1000 for sklearn 0.21.
                    {'enum': [None]}],
                'default': None,
                'description': 'The maximum number of passes over the training data (aka epochs).'},
            'tol': {
                'anyOf': [{
                    'type': 'number',
                    'minimumForOptimizer': 1e-08,
                    'maximumForOptimizer': 0.01,
                    'distribution': 'loguniform'}, {
                    'enum': [None]}],
                'default': None, #default value is 1e-3 from sklearn 0.21.
                'description': 'The stopping criterion. If it is not None, the iterations will stop'},
            'early_stopping': {
                'type': 'boolean',
                'default': False,
                'description': 'Whether to use early stopping to terminate training when validation.'},
            'validation_fraction': {
                'type': 'number',
                'default': 0.1,
                'description': 'The proportion of training data to set aside as validation set for'},
            'n_iter_no_change': {
                'type': 'integer',
                'minimumForOptimizer': 5,
                'maximumForOptimizer': 10,
                'default': 5,
                'description': 'Number of iterations with no improvement to wait before early stopping.'},
            'shuffle': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether or not the training data should be shuffled after each epoch.'},
            'verbose': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': 0,
                'description': 'The verbosity level'},
            'loss': {
                'enum': ['hinge', 'squared_hinge'],
                'default': 'hinge',
                'description': 'The loss function to be used:'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The number of CPUs to use to do the OVA (One Versus All, for'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The seed of the pseudo random number generator to use when shuffling'},
            'warm_start': {
                'type': 'boolean',
                'default': False,
                'description': 'When set to True, reuse the solution of the previous call to' 
                ' fit as initialization, otherwise, just erase the previous solution.'},
            'class_weight': {
                'anyOf': [{
                    'type': 'object'}, {
                    'enum': ['balanced', None]}],
                'default': None,
                'description': 'Preset for the class_weight fit parameter.'},
            'average': {
                'anyOf': [{
                    'type': 'boolean'}, {
                    'type': 'integer',
                    'forOptimizer': False}],
                'default': False,
                'description': 'When set to True, computes the averaged SGD weights and stores the'}
        }},
        {'description': 'validation_fraction, only used if early_stopping is true',
        'anyOf': [{
            'type': 'object',
            'properties': {
                'early_stopping': {
                    'enum': [True]},
            }}, {
            'type': 'object',
            'properties': {
                'validation_fraction': {
                    'enum': [0.1]}, #i.e. it should not have a value other than its default.
            }}]}]}

_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit linear model with Passive Aggressive algorithm.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'description': 'Training data',
            'type': 'array',
            'items': {
                'type': 'array',
                'items': { 'type': 'number'}}},
        'y': {
            'description': 'Target values',
            'anyOf': [
                {'type': 'array', 'items': {'type': 'number'}},
                {'type': 'array', 'items': {'type': 'string'}}]
            },
        'coef_init': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': { 'type': 'number'}},
            'description': 'The initial coefficients to warm-start the optimization.'},
        'intercept_init': {
            'type': 'array',
            'items': {
                    'type': 'number'},
            'description': 'The initial intercept to warm-start the optimization.'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict class labels for samples in X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'description': 'Test data',
            'type': 'array',
            'items': {
                'type': 'array',
                'items': { 'type': 'number'}}},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict class labels for samples in X.',
    'anyOf': [
        {'type': 'array', 'items': {'type': 'number'}},
        {'type': 'array', 'items': {'type': 'string'}}]}

_input_decision_function_schema = {
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}}}}

_output_decision_function_schema = {
    'description': 'Confidence scores for samples for each class in the model.',
    'anyOf': [
    {   'description': 'In the multi-way case, score per (sample, class) combination.',
        'type': 'array',
        'items': {'type': 'array', 'items': {'type': 'number'}}},
    {   'description': 'In the binary case, score for `self._classes[1]`.',
        'type': 'array',
        'items': {'type': 'number'}}]}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.PassiveAggressiveClassifier.html#sklearn-linear-model-passiveaggressiveclassifier',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'classifier'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_decision_function': _input_decision_function_schema,
        'output_decision_function': _output_decision_function_schema,
}}

lale.docstrings.set_docstrings(PassiveAggressiveClassifierImpl, _combined_schemas)

PassiveAggressiveClassifier = lale.operators.make_operator(PassiveAggressiveClassifierImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import warnings
import random
import sys
import lale.operators as Ops
from lale.lib.lale import ConcatFeatures
from lale.lib.lale import NoOp
from lale.lib.sklearn import KNeighborsClassifier
from lale.lib.sklearn import LinearSVC
from lale.lib.sklearn import LogisticRegression
from lale.lib.sklearn import MinMaxScaler
from lale.lib.sklearn import MLPClassifier
from lale.lib.sklearn import Nystroem
from lale.lib.sklearn import OneHotEncoder
from lale.lib.sklearn import PCA
from lale.lib.sklearn import TfidfVectorizer
from lale.lib.sklearn import MultinomialNB
from lale.lib.sklearn import SimpleImputer
from lale.lib.sklearn import SVC
from lale.lib.xgboost import XGBClassifier
from lale.lib.sklearn import PassiveAggressiveClassifier
from lale.lib.sklearn import StandardScaler
from lale.lib.sklearn import FeatureAgglomeration
from typing import List
import lale.type_checking

import sklearn.datasets

from lale.sklearn_compat import make_sklearn_compat
from lale.search.lale_smac import get_smac_space, lale_trainable_op_from_config
from lale.search.op2hp import hyperopt_search_space

@unittest.skip("Skipping here because travis-ci fails to allocate memory. This runs on internal travis.")
class TestResNet50(unittest.TestCase):
    def test_init_fit_predict(self):
        import torchvision.datasets as datasets
        import torchvision.transforms as transforms
        from lale.lib.pytorch import ResNet50

        transform = transforms.Compose([transforms.ToTensor()])

        data_train = datasets.FakeData(size = 50, num_classes=2 , transform = transform)#, target_transform = transform)
        clf = ResNet50(num_classes=2,num_epochs = 1)
        clf.fit(data_train)
        predicted = clf.predict(data_train)


class TestResamplers(unittest.TestCase):

    def setUp(self):
        from sklearn.datasets import make_classification
        from sklearn.model_selection import train_test_split
        X, y = make_classification(n_classes=2, class_sep=2,
            weights=[0.1, 0.9], n_informative=3, n_redundant=1, flip_y=0,
            n_features=20, n_clusters_per_class=1, n_samples=1000, random_state=10)
        self.X_train, self.X_test, self.y_train, self.y_test =  train_test_split(X, y)    

def create_function_test_resampler(res_name):
    def test_resampler(self):
        from lale.lib.sklearn import PCA, Nystroem, LogisticRegression, RandomForestClassifier
        from lale.lib.lale import NoOp, ConcatFeatures
        X_train, y_train = self.X_train, self.y_train
        X_test, y_test = self.X_test, self.y_test
        import importlib
        module_name = ".".join(res_name.split('.')[0:-1])
        class_name = res_name.split('.')[-1]
        module = importlib.import_module(module_name)

        class_ = getattr(module, class_name)
        with self.assertRaises(ValueError):
            res = class_()

        #test_schemas_are_schemas
        lale.type_checking.validate_is_schema(class_.input_schema_fit())
        lale.type_checking.validate_is_schema(class_.input_schema_predict())
        lale.type_checking.validate_is_schema(class_.output_schema_predict())
        lale.type_checking.validate_is_schema(class_.hyperparam_schema())

        #test_init_fit_predict
        from lale.operators import make_pipeline
        pipeline1 = PCA() >> class_(operator=make_pipeline(LogisticRegression()))
        trained = pipeline1.fit(X_train, y_train)
        predictions = trained.predict(X_test)

        pipeline2 = class_(operator=make_pipeline(PCA(), LogisticRegression()))
        trained = pipeline2.fit(X_train, y_train)
        predictions = trained.predict(X_test)

        #test_with_hyperopt
        from lale.lib.lale import Hyperopt
        optimizer = Hyperopt(estimator=PCA >> class_(operator=make_pipeline(LogisticRegression())), max_evals = 1, show_progressbar=False)
        trained_optimizer = optimizer.fit(X_train, y_train)
        predictions = trained_optimizer.predict(X_test)

        pipeline3 = class_(operator= PCA() >> (Nystroem & NoOp) >> ConcatFeatures >> LogisticRegression())
        optimizer = Hyperopt(estimator=pipeline3, max_evals = 1, show_progressbar=False)
        trained_optimizer = optimizer.fit(X_train, y_train)
        predictions = trained_optimizer.predict(X_test)

        pipeline4 = (PCA >> class_(operator=make_pipeline(Nystroem())) & class_(operator=make_pipeline(Nystroem()))) >> ConcatFeatures >> LogisticRegression()
        optimizer = Hyperopt(estimator=pipeline4, max_evals = 1, scoring='roc_auc', show_progressbar=False)
        trained_optimizer = optimizer.fit(X_train, y_train)
        predictions = trained_optimizer.predict(X_test)

        #test_cross_validation
        from lale.helpers import cross_val_score
        cv_results = cross_val_score(pipeline1, X_train, y_train, cv = 2)
        self.assertEqual(len(cv_results), 2)

        #test_to_json
        pipeline1.to_json()

    test_resampler.__name__ = 'test_{0}'.format(res_name.split('.')[-1])
    return test_resampler

resamplers = ['lale.lib.imblearn.SMOTE',
              'lale.lib.imblearn.SMOTEENN',
              'lale.lib.imblearn.ADASYN',
              'lale.lib.imblearn.BorderlineSMOTE',
              'lale.lib.imblearn.SVMSMOTE',
              'lale.lib.imblearn.RandomOverSampler',
              'lale.lib.imblearn.CondensedNearestNeighbour',
              'lale.lib.imblearn.EditedNearestNeighbours',
              'lale.lib.imblearn.RepeatedEditedNearestNeighbours',
              'lale.lib.imblearn.AllKNN',
              'lale.lib.imblearn.InstanceHardnessThreshold']

for res in resamplers:
    setattr(
        TestResamplers,
        'test_{0}'.format(res.split('.')[-1]),
        create_function_test_resampler(res)
    )

class TestImblearn(unittest.TestCase):
    def setUp(self):
        from sklearn.datasets import make_classification
        from sklearn.model_selection import train_test_split
        X, y = make_classification(n_classes=2, class_sep=2,
            weights=[0.1, 0.9], n_informative=3, n_redundant=1, flip_y=0,
            n_features=20, n_clusters_per_class=1, n_samples=1000, random_state=10)
        self.X_train, self.X_test, self.y_train, self.y_test =  train_test_split(X, y)    

    def test_decision_function(self):
        from lale.lib.imblearn import SMOTE
        from lale.operators import make_pipeline
        from lale.lib.sklearn import RandomForestClassifier
        smote = SMOTE(operator=make_pipeline(RandomForestClassifier()))
        trained = smote.fit(self.X_train, self.y_train)
        trained.predict(self.X_test)
        with self.assertRaises(AttributeError):
            trained.decision_function(self.X_test)
    
    def test_string_labels(self):
        from lale.lib.imblearn import CondensedNearestNeighbour
        print(type(CondensedNearestNeighbour))
        from lale.operators import make_pipeline
        y_train = ['low' if label==0 else 'high' for label in self.y_train]
        pipeline = CondensedNearestNeighbour(operator=make_pipeline(PCA(), LogisticRegression()), sampling_strategy=['high'])
        trained = pipeline.fit(self.X_train, y_train)
        predictions = trained.predict(self.X_test)




from sklearn.compose._target import TransformedTargetRegressor as SKLModel
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class TransformedTargetRegressorImpl():

    def __init__(self, regressor=None, transformer=None, func=None, inverse_func=None, check_inverse=True):
        self._hyperparams = {
            'regressor': regressor,
            'transformer': transformer,
            'func': func,
            'inverse_func': inverse_func,
            'check_inverse': check_inverse}
        self._wrapped_model = SKLModel(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for TransformedTargetRegressor    Meta-estimator to regress on a transformed target.',
    'allOf': [{
        'type': 'object',
        'required': ['regressor', 'transformer', 'func', 'inverse_func', 'check_inverse'],
        'relevantToOptimizer': [],
        'additionalProperties': False,
        'properties': {
            'regressor': {
                'XXX TODO XXX': 'object, default=LinearRegression()',
                'description': 'Regressor object such as derived from ``RegressorMixin``',
                'enum': [None],
                'default': None},
            'transformer': {
                'anyOf': [{
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Estimator object such as derived from ``TransformerMixin``'},
            'func': {
                'XXX TODO XXX': 'function, optional',
                'description': 'Function to apply to ``y`` before passing to ``fit``',
                'enum': [None],
                'default': None},
            'inverse_func': {
                'XXX TODO XXX': 'function, optional',
                'description': 'Function to apply to the prediction of the regressor',
                'enum': [None],
                'default': None},
            'check_inverse': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether to check that ``transform`` followed by ``inverse_transform`` or ``func`` followed by ``inverse_func`` leads to the original targets.'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the model according to the given training data.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vector, where n_samples is the number of samples and n_features is the number of features.'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Target values.'},
        'sample_weight': {
            'laleType': 'Any',
            'XXX TODO XXX': 'array-like, shape (n_samples,) optional',
            'description': 'Array of weights that are assigned to individual samples'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict using the base regressor, applying inverse.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Samples.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predicted values.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.compose.TransformedTargetRegressor#sklearn-compose-transformedtargetregressor',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'regressor'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema},
}
lale.docstrings.set_docstrings(TransformedTargetRegressorImpl, _combined_schemas)
TransformedTargetRegressor = lale.operators.make_operator(TransformedTargetRegressorImpl, _combined_schemas)

